# -*- coding: utf-8 -*-

# Weechat Matrix Protocol Script
# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import unicode_literals

import time
import attr
import pprint
from builtins import super
from functools import partial
from collections import deque
from typing import Dict, List, NamedTuple, Optional, Set
from uuid import UUID

from nio import (
    Api,
    PowerLevelsEvent,
    RedactedEvent,
    RedactionEvent,
    RoomAliasEvent,
    RoomEncryptionEvent,
    RoomMemberEvent,
    RoomMessage,
    RoomMessageEmote,
    RoomMessageMedia,
    RoomEncryptedMedia,
    RoomMessageNotice,
    RoomMessageText,
    RoomMessageUnknown,
    RoomNameEvent,
    RoomTopicEvent,
    MegolmEvent,
    Event,
    OlmTrustError,
    UnknownEvent,
    FullyReadEvent,
    BadEvent,
    UnknownBadEvent,
)

from . import globals as G
from .colors import Formatted
from .config import RedactType, NewChannelPosition
from .globals import SCRIPT_NAME, SERVERS, W, TYPING_NOTICE_TIMEOUT
from .utf import utf8_decode
from .message_renderer import Render
from .utils import (
    server_ts_to_weechat,
    shorten_sender,
    string_strikethrough,
    color_pair,
)


@attr.s
class OwnMessages(object):
    sender = attr.ib(type=str)
    age = attr.ib(type=int)
    event_id = attr.ib(type=str)
    uuid = attr.ib(type=str)
    room_id = attr.ib(type=str)
    formatted_message = attr.ib(type=Formatted)


class OwnMessage(OwnMessages):
    pass


class OwnAction(OwnMessage):
    pass


@utf8_decode
def room_buffer_input_cb(server_name, buffer, input_data):
    server = SERVERS[server_name]
    room_buffer = server.find_room_from_ptr(buffer)

    if not room_buffer:
        # TODO log error
        return W.WEECHAT_RC_ERROR

    if not server.connected:
        room_buffer.error("You are not connected to the server")
        return W.WEECHAT_RC_ERROR

    if not server.client.logged_in:
        room_buffer.error("You are not logged in.")
        return W.WEECHAT_RC_ERROR

    data = W.string_input_for_buffer(input_data)

    if not data:
        data = input_data

    formatted_data = Formatted.from_input_line(data)

    try:
        server.room_send_message(room_buffer, formatted_data, "m.text")
        room_buffer.last_message = None
    except OlmTrustError as e:
        if (G.CONFIG.network.resending_ignores_devices
                and room_buffer.last_message):
            room_buffer.error("Ignoring unverified devices.")

            if (room_buffer.last_message.to_weechat() ==
                    formatted_data.to_weechat()):
                server.room_send_message(room_buffer, formatted_data, "m.text",
                                         ignore_unverified_devices=True)
                room_buffer.last_message = None
        else:
            # If the item is a normal user message store it in the
            # buffer to enable the send-anyways functionality.
            room_buffer.error("Untrusted devices found in room: {}".format(e))
            room_buffer.last_message = formatted_data

    return W.WEECHAT_RC_OK


@utf8_decode
def room_buffer_close_cb(server_name, buffer):
    server = SERVERS[server_name]
    room_buffer = server.find_room_from_ptr(buffer)

    if room_buffer:
        room_id = room_buffer.room.room_id
        server.buffers.pop(room_id, None)
        server.room_buffers.pop(room_id, None)

    return W.WEECHAT_RC_OK


class WeechatUser(object):
    def __init__(self, nick, host=None, prefix="", join_time=None):
        # type: (str, str, str, int) -> None
        self.nick = nick
        self.host = host
        self.prefix = prefix
        self.color = W.info_get("nick_color_name", nick)
        self.join_time = join_time or time.time()
        self.speaking_time = None  # type: Optional[int]

    def update_speaking_time(self, new_time=None):
        self.speaking_time = new_time or time.time()

    @property
    def joined_recently(self):
        # TODO make the delay configurable
        delay = 30
        limit = time.time() - (delay * 60)
        return self.join_time < limit

    @property
    def spoken_recently(self):
        if not self.speaking_time:
            return False

        # TODO make the delay configurable
        delay = 5
        limit = time.time() - (delay * 60)
        return self.speaking_time < limit


class RoomUser(WeechatUser):
    def __init__(self, nick, user_id=None, power_level=0, join_time=None):
        # type: (str, str, int, int) -> None
        prefix = self._get_prefix(power_level)
        super().__init__(nick, user_id, prefix, join_time)

    @property
    def power_level(self):
        # This shouldn't be used since it's a lossy function. It's only here
        # for the setter
        if self.prefix == "&":
            return 100
        if self.prefix == "@":
            return 50
        if self.prefix == "+":
            return 10
        return 0

    @power_level.setter
    def power_level(self, level):
        self.prefix = self._get_prefix(level)

    @staticmethod
    def _get_prefix(power_level):
        # type: (int) -> str
        if power_level >= 100:
            return "&"
        if power_level >= 50:
            return "@"
        if power_level > 0:
            return "+"
        return ""


class WeechatChannelBuffer(object):
    tags = {
        "message": [SCRIPT_NAME + "_message", "notify_message", "log1"],
        "message_private": [
            SCRIPT_NAME + "_message",
            "notify_private",
            "log1"
        ],
        "self_message": [
            SCRIPT_NAME + "_message",
            "notify_none",
            "no_highlight",
            "self_msg",
            "log1",
        ],
        "action": [
            SCRIPT_NAME + "_message",
            SCRIPT_NAME + "_action",
            "notify_message",
            "log1",
        ],
        "action_private": [
            SCRIPT_NAME + "_message",
            SCRIPT_NAME + "_action",
            "notify_private",
            "log1",
        ],
        "notice": [SCRIPT_NAME + "_notice", "notify_message", "log1"],
        "old_message": [
            SCRIPT_NAME + "_message",
            "notify_message",
            "no_log",
            "no_highlight",
        ],
        "join": [SCRIPT_NAME + "_join", "log4"],
        "part": [SCRIPT_NAME + "_leave", "log4"],
        "kick": [SCRIPT_NAME + "_kick", "log4"],
        "invite": [SCRIPT_NAME + "_invite", "log4"],
        "topic": [SCRIPT_NAME + "_topic", "log3"],
    }

    membership_messages = {
        "join": "has joined",
        "part": "has left",
        "kick": "has been kicked from",
        "invite": "has been invited to",
    }

    class Line(object):
        def __init__(self, pointer):
            self._ptr = pointer

        @property
        def _hdata(self):
            return W.hdata_get("line_data")

        @property
        def prefix(self):
            return W.hdata_string(self._hdata, self._ptr, "prefix")

        @prefix.setter
        def prefix(self, new_prefix):
            new_data = {"prefix": new_prefix}
            W.hdata_update(self._hdata, self._ptr, new_data)

        @property
        def message(self):
            return W.hdata_string(self._hdata, self._ptr, "message")

        @message.setter
        def message(self, new_message):
            # type: (str) -> None
            new_data = {"message": new_message}
            W.hdata_update(self._hdata, self._ptr, new_data)

        @property
        def tags(self):
            tags_count = W.hdata_get_var_array_size(
                self._hdata, self._ptr, "tags_array"
            )

            tags = [
                W.hdata_string(self._hdata, self._ptr, "%d|tags_array" % i)
                for i in range(tags_count)
            ]
            return tags

        @tags.setter
        def tags(self, new_tags):
            # type: (List[str]) -> None
            new_data = {"tags_array": ",".join(new_tags)}
            W.hdata_update(self._hdata, self._ptr, new_data)

        @property
        def date(self):
            # type: () -> int
            return W.hdata_time(self._hdata, self._ptr, "date")

        @date.setter
        def date(self, new_date):
            # type: (int) -> None
            new_data = {"date": str(new_date)}
            W.hdata_update(self._hdata, self._ptr, new_data)

        @property
        def date_printed(self):
            # type: () -> int
            return W.hdata_time(self._hdata, self._ptr, "date_printed")

        @date_printed.setter
        def date_printed(self, new_date):
            # type: (int) -> None
            new_data = {"date_printed": str(new_date)}
            W.hdata_update(self._hdata, self._ptr, new_data)

        @property
        def highlight(self):
            # type: () -> bool
            return bool(W.hdata_char(self._hdata, self._ptr, "highlight"))

        def update(
            self,
            date=None,
            date_printed=None,
            tags=None,
            prefix=None,
            message=None,
            highlight=None,
        ):
            new_data = {}

            if date is not None:
                new_data["date"] = str(date)
            if date_printed is not None:
                new_data["date_printed"] = str(date_printed)
            if tags is not None:
                new_data["tags_array"] = ",".join(tags)
            if prefix is not None:
                new_data["prefix"] = prefix
            if message is not None:
                new_data["message"] = message
            if highlight is not None:
                new_data["highlight"] = highlight

            if new_data:
                W.hdata_update(self._hdata, self._ptr, new_data)

    def __init__(self, name, server_name, user):
        # type: (str, str, str) -> None

        # Previous buffer num before create
        cur_num = W.buffer_get_integer(W.current_buffer(), "number")
        self._ptr = W.buffer_new(
            name,
            "room_buffer_input_cb",
            server_name,
            "room_buffer_close_cb",
            server_name,
        )

        new_channel_position = G.CONFIG.look.new_channel_position
        if new_channel_position == NewChannelPosition.NONE:
            pass
        elif new_channel_position == NewChannelPosition.NEXT:
            self.number = cur_num + 1
        elif new_channel_position == NewChannelPosition.NEAR_SERVER:
            server = G.SERVERS[server_name]
            last_similar_buffer_num = max(
                (room.weechat_buffer.number for room
                    in server.room_buffers.values()),
                default=W.buffer_get_integer(server.server_buffer, "number")
            )
            self.number = last_similar_buffer_num + 1

        self.name = ""
        self.users = {}  # type: Dict[str, WeechatUser]
        self.smart_filtered_nicks = set()  # type: Set[str]

        self.topic_author = ""
        self.topic_date = None

        W.buffer_set(self._ptr, "localvar_set_type", "private")
        W.buffer_set(self._ptr, "type", "formatted")

        W.buffer_set(self._ptr, "localvar_set_channel", name)

        W.buffer_set(self._ptr, "localvar_set_nick", user)

        W.buffer_set(self._ptr, "localvar_set_server", server_name)

        W.nicklist_add_group(
            self._ptr, "", "000|o", "weechat.color.nicklist_group", 1
        )
        W.nicklist_add_group(
            self._ptr, "", "001|h", "weechat.color.nicklist_group", 1
        )
        W.nicklist_add_group(
            self._ptr, "", "002|v", "weechat.color.nicklist_group", 1
        )
        W.nicklist_add_group(
            self._ptr, "", "999|...", "weechat.color.nicklist_group", 1
        )

        W.buffer_set(self._ptr, "nicklist", "1")
        W.buffer_set(self._ptr, "nicklist_display_groups", "0")

        W.buffer_set(self._ptr, "highlight_words", user)

        # TODO make this configurable
        W.buffer_set(
            self._ptr, "highlight_tags_restrict", SCRIPT_NAME + "_message"
        )

    @property
    def _hdata(self):
        return W.hdata_get("buffer")

    def add_smart_filtered_nick(self, nick):
        self.smart_filtered_nicks.add(nick)

    def remove_smart_filtered_nick(self, nick):
        self.smart_filtered_nicks.discard(nick)

    def unmask_smart_filtered_nick(self, nick):
        if nick not in self.smart_filtered_nicks:
            return

        for line in self.lines:
            filtered = False
            join = False
            tags = line.tags

            if "nick_{}".format(nick) not in tags:
                continue

            if SCRIPT_NAME + "_smart_filter" in tags:
                filtered = True
            elif SCRIPT_NAME + "_join" in tags:
                join = True

            if filtered:
                tags.remove(SCRIPT_NAME + "_smart_filter")
                line.tags = tags

            if join:
                break

        self.remove_smart_filtered_nick(nick)

    @property
    def input(self):
        # type: () -> str
        """Get the bar item input text of the buffer."""
        return W.buffer_get_string(self._ptr, "input")

    @property
    def num_lines(self):
        own_lines = W.hdata_pointer(self._hdata, self._ptr, "own_lines")
        return W.hdata_integer(W.hdata_get("lines"), own_lines, "lines_count")

    @property
    def lines(self):
        own_lines = W.hdata_pointer(self._hdata, self._ptr, "own_lines")

        if own_lines:
            hdata_line = W.hdata_get("line")

            line_pointer = W.hdata_pointer(
                W.hdata_get("lines"), own_lines, "last_line"
            )

            while line_pointer:
                data_pointer = W.hdata_pointer(
                    hdata_line, line_pointer, "data"
                )

                if data_pointer:
                    yield WeechatChannelBuffer.Line(data_pointer)

                line_pointer = W.hdata_move(hdata_line, line_pointer, -1)

    def _print(self, string):
        # type: (str) -> None
        """ Print a string to the room buffer """
        W.prnt(self._ptr, string)

    def print_date_tags(self, data, date=None, tags=None):
        # type: (str, Optional[int], Optional[List[str]]) -> None
        date = date or int(time.time())
        tags = tags or []

        tags_string = ",".join(tags)
        W.prnt_date_tags(self._ptr, date, tags_string, data)

    def error(self, string):
        # type: (str) -> None
        """ Print an error to the room buffer """
        message = "{prefix}{script}: {message}".format(
            prefix=W.prefix("error"), script=SCRIPT_NAME, message=string
        )

        self._print(message)

    def info(self, string):
        message = "{prefix}{script}: {message}".format(
            prefix=W.prefix("network"), script=SCRIPT_NAME, message=string
        )
        self._print(message)

    @staticmethod
    def _color_for_tags(color):
        # type: (str) -> str
        if color == "weechat.color.chat_nick_self":
            option = W.config_get(color)
            return W.config_string(option)

        return color

    def _message_tags(self, user, message_type):
        # type: (WeechatUser, str) -> List[str]
        tags = list(self.tags[message_type])

        tags.append("nick_{nick}".format(nick=user.nick))

        color = self._color_for_tags(user.color)

        if message_type not in ("action", "notice"):
            tags.append("prefix_nick_{color}".format(color=color))

        return tags

    def _get_user(self, nick):
        # type: (str) -> WeechatUser
        if nick in self.users:
            return self.users[nick]

        # A message from a non joined user
        return WeechatUser(nick)

    def _print_message(self, user, message, date, tags, extra_prefix=""):
        prefix_string = (
            extra_prefix
            if not user.prefix
            else "{}{}{}{}".format(
                extra_prefix,
                W.color(self._get_prefix_color(user.prefix)),
                user.prefix,
                W.color("reset"),
            )
        )

        data = "{prefix}{color}{author}{ncolor}\t{msg}".format(
            prefix=prefix_string,
            color=W.color(user.color),
            author=user.nick,
            ncolor=W.color("reset"),
            msg=message,
        )

        self.print_date_tags(data, date, tags)

    def message(self, nick, message, date, extra_tags=None, extra_prefix=""):
        # type: (str, str, int, List[str], str) -> None
        user = self._get_user(nick)
        tags_type = "message_private" if self.type == "private" else "message"
        tags = self._message_tags(user, tags_type) + (extra_tags or [])
        self._print_message(user, message, date, tags, extra_prefix)

        user.update_speaking_time(date)
        self.unmask_smart_filtered_nick(nick)

    def notice(self, nick, message, date, extra_tags=None, extra_prefix=""):
        # type: (str, str, int, Optional[List[str]], str) -> None
        user = self._get_user(nick)
        user_prefix = (
            ""
            if not user.prefix
            else "{}{}{}".format(
                W.color(self._get_prefix_color(user.prefix)),
                user.prefix,
                W.color("reset"),
            )
        )

        user_string = "{}{}{}{}".format(
            user_prefix, W.color(user.color), user.nick, W.color("reset")
        )

        data = (
            "{extra_prefix}{prefix}{color}Notice"
            "{del_color}({ncolor}{user}{del_color}){ncolor}"
            ": {message}"
        ).format(
            extra_prefix=extra_prefix,
            prefix=W.prefix("network"),
            color=W.color("irc.color.notice"),
            del_color=W.color("chat_delimiters"),
            ncolor=W.color("reset"),
            user=user_string,
            message=message,
        )

        tags = self._message_tags(user, "notice") + (extra_tags or [])
        self.print_date_tags(data, date, tags)

        user.update_speaking_time(date)
        self.unmask_smart_filtered_nick(nick)

    def _format_action(self, user, message):
        nick_prefix = (
            ""
            if not user.prefix
            else "{}{}{}".format(
                W.color(self._get_prefix_color(user.prefix)),
                user.prefix,
                W.color("reset"),
            )
        )

        data = (
            "{nick_prefix}{nick_color}{author}"
            "{ncolor} {msg}").format(
            nick_prefix=nick_prefix,
            nick_color=W.color(user.color),
            author=user.nick,
            ncolor=W.color("reset"),
            msg=message,
        )
        return data

    def _print_action(self, user, message, date, tags, extra_prefix=""):
        data = self._format_action(user, message)
        data = "{extra_prefix}{prefix}{data}".format(
            extra_prefix=extra_prefix,
            prefix=W.prefix("action"),
            data=data)

        self.print_date_tags(data, date, tags)

    def action(self, nick, message, date, extra_tags=None, extra_prefix=""):
        # type: (str, str, int, Optional[List[str]], str) -> None
        user = self._get_user(nick)
        tags_type = "action_private" if self.type == "private" else "action"
        tags = self._message_tags(user, tags_type) + (extra_tags or [])
        self._print_action(user, message, date, tags, extra_prefix)

        user.update_speaking_time(date)
        self.unmask_smart_filtered_nick(nick)

    @staticmethod
    def _get_nicklist_group(user):
        # type: (WeechatUser) -> str
        group_name = "999|..."

        if user.prefix == "&":
            group_name = "000|o"
        elif user.prefix == "@":
            group_name = "001|h"
        elif user.prefix == "+":
            group_name = "002|v"

        return group_name

    @staticmethod
    def _get_prefix_color(prefix):
        # type: (str) -> str

        return G.CONFIG.color.nick_prefixes.get(prefix, "")

    def _add_user_to_nicklist(self, user):
        # type: (WeechatUser) -> None
        nick_pointer = W.nicklist_search_nick(self._ptr, "", user.nick)

        if not nick_pointer:
            group = W.nicklist_search_group(
                self._ptr, "", self._get_nicklist_group(user)
            )
            prefix = user.prefix if user.prefix else " "
            W.nicklist_add_nick(
                self._ptr,
                group,
                user.nick,
                user.color,
                prefix,
                self._get_prefix_color(user.prefix),
                1,
            )

    def _membership_message(self, user, message_type):
        # type: (WeechatUser, str) -> str
        action_color = "green" if message_type in ("join", "invite") else "red"
        prefix = "join" if message_type in ("join", "invite") else "quit"

        membership_message = self.membership_messages[message_type]

        message = (
            "{prefix}{color}{author}{ncolor} "
            "{del_color}({host_color}{host}{del_color})"
            "{action_color} {message} "
            "{channel_color}{room}{ncolor}"
        ).format(
            prefix=W.prefix(prefix),
            color=W.color(user.color),
            author=user.nick,
            ncolor=W.color("reset"),
            del_color=W.color("chat_delimiters"),
            host_color=W.color("chat_host"),
            host=user.host,
            action_color=W.color(action_color),
            message=membership_message,
            channel_color=W.color("chat_channel"),
            room=self.short_name,
        )

        return message

    def join(self, user, date, message=True, extra_tags=None):
        # type: (WeechatUser, int, Optional[bool], Optional[List[str]]) -> None
        self._add_user_to_nicklist(user)
        self.users[user.nick] = user

        if len(self.users) > 2:
            W.buffer_set(self._ptr, "localvar_set_type", "channel")

        if message:
            tags = self._message_tags(user, "join")
            msg = self._membership_message(user, "join")

            # TODO add a option to disable smart filters
            tags.append(SCRIPT_NAME + "_smart_filter")

            self.print_date_tags(msg, date, tags)
            self.add_smart_filtered_nick(user.nick)

    def invite(self, nick, date, extra_tags=None):
        # type: (str, int, Optional[List[str]]) -> None
        user = self._get_user(nick)
        tags = self._message_tags(user, "invite")
        message = self._membership_message(user, "invite")
        self.print_date_tags(message, date, tags + (extra_tags or []))

    def remove_user_from_nicklist(self, user):
        # type: (WeechatUser) -> None
        nick_pointer = W.nicklist_search_nick(self._ptr, "", user.nick)

        if nick_pointer:
            W.nicklist_remove_nick(self._ptr, nick_pointer)

    def _leave(self, nick, date, message, leave_type, extra_tags=None):
        # type: (str, int, bool, str, List[str]) -> None
        user = self._get_user(nick)
        self.remove_user_from_nicklist(user)

        if len(self.users) <= 2:
            W.buffer_set(self._ptr, "localvar_set_type", "private")

        if message:
            tags = self._message_tags(user, leave_type)

            # TODO make this configurable
            if not user.spoken_recently:
                tags.append(SCRIPT_NAME + "_smart_filter")

            msg = self._membership_message(user, leave_type)
            self.print_date_tags(msg, date, tags + (extra_tags or []))
            self.remove_smart_filtered_nick(user.nick)

        if user.nick in self.users:
            del self.users[user.nick]

    def part(self, nick, date, message=True, extra_tags=None):
        # type: (str, int, bool, Optional[List[str]]) -> None
        self._leave(nick, date, message, "part", extra_tags)

    def kick(self, nick, date, message=True, extra_tags=None):
        # type: (str, int, bool, Optional[List[str]]) -> None
        self._leave(nick, date, message, "kick", extra_tags)

    def _print_topic(self, nick, topic, date):
        user = self._get_user(nick)
        tags = self._message_tags(user, "topic")

        data = (
            "{prefix}{nick} has changed "
            "the topic for {chan_color}{room}{ncolor} "
            'to "{topic}"'
        ).format(
            prefix=W.prefix("network"),
            nick=user.nick,
            chan_color=W.color("chat_channel"),
            ncolor=W.color("reset"),
            room=self.short_name,
            topic=topic,
        )

        self.print_date_tags(data, date, tags)
        user.update_speaking_time(date)
        self.unmask_smart_filtered_nick(nick)

    @property
    def topic(self):
        return W.buffer_get_string(self._ptr, "title")

    @topic.setter
    def topic(self, topic):
        W.buffer_set(self._ptr, "title", topic)

    def change_topic(self, nick, topic, date, message=True):
        if message:
            self._print_topic(nick, topic, date)

        self.topic = topic
        self.topic_author = nick
        self.topic_date = date

    def self_message(self, nick, message, date, tags=None):
        user = self._get_user(nick)
        tags = self._message_tags(user, "self_message") + (tags or [])
        self._print_message(user, message, date, tags)

    def self_action(self, nick, message, date, tags=None):
        user = self._get_user(nick)
        tags = self._message_tags(user, "self_message") + (tags or [])
        tags.append(SCRIPT_NAME + "_action")
        self._print_action(user, message, date, tags)

    @property
    def type(self):
        return W.buffer_get_string(self._ptr, "localvar_type")

    @property
    def short_name(self):
        return W.buffer_get_string(self._ptr, "short_name")

    @short_name.setter
    def short_name(self, name):
        W.buffer_set(self._ptr, "short_name", name)

    @property
    def name(self):
        return W.buffer_get_string(self._ptr, "name")

    @name.setter
    def name(self, name):
        W.buffer_set(self._ptr, "name", name)

    @property
    def number(self):
        """Get the buffer number, starts at 1."""
        return int(W.buffer_get_integer(self._ptr, "number"))

    @number.setter
    def number(self, n):
        W.buffer_set(self._ptr, "number", str(n))

    def find_lines(self, predicate, max_lines=None):
        lines = []
        count = 0
        for line in self.lines:
            if predicate(line):
                lines.append(line)
                count += 1
                if max_lines is not None and count == max_lines:
                    return lines

        return lines


class RoomBuffer(object):
    def __init__(self, room, server_name, homeserver, prev_batch):
        self.room = room
        self.homeserver = homeserver
        self._backlog_pending = False
        self.prev_batch = prev_batch
        self.joined = True
        self.leave_event_id = None  # type: Optional[str]
        self.members_fetched = False
        self.first_view = True
        self.first_backlog_request = True
        self.unhandled_users = []   # type: List[str]
        self.inactive_users = []

        self.sent_messages_queue = dict()  # type: Dict[UUID, OwnMessage]
        self.printed_before_ack_queue = list()  # type: List[UUID]
        self.undecrypted_events = deque(maxlen=5000)

        self.typing_notice_time = None
        self._typing = False
        self.typing_enabled = True

        self.last_read_event = None
        self._read_markers_enabled = True
        self.server_name = server_name

        self.last_message = None

        buffer_name = "{}{}.{}".format(G.BUFFER_NAME_PREFIX, server_name, room.room_id)

        # This dict remembers the connection from a user_id to the name we
        # displayed in the buffer
        self.displayed_nicks = {}
        user = shorten_sender(self.room.own_user_id)

        self.weechat_buffer = WeechatChannelBuffer(
            buffer_name, server_name, user
        )

        W.buffer_set(
            self.weechat_buffer._ptr,
            "localvar_set_domain",
            self.homeserver.hostname
        )

        W.buffer_set(
            self.weechat_buffer._ptr,
            "localvar_set_room_id",
            room.room_id
        )

        if room.canonical_alias:
            self.update_canonical_alias_localvar()

    @property
    def backlog_pending(self):
        return self._backlog_pending

    @backlog_pending.setter
    def backlog_pending(self, value):
        self._backlog_pending = value
        W.bar_item_update("buffer_modes")
        W.bar_item_update("matrix_modes")

    @property
    def warning_prefix(self):
        return G.CONFIG.look.encryption_warning_sign

    @property
    def typing(self):
        # type: () -> bool
        """Return our typing status."""
        return self._typing

    @typing.setter
    def typing(self, value):
        self._typing = value
        if value:
            self.typing_notice_time = time.time()
        else:
            self.typing_notice_time = None

    @property
    def typing_notice_expired(self):
        # type: () -> bool
        """Check if the typing notice has expired.

        Returns true if a new typing notice should be sent.
        """
        if not self.typing_notice_time:
            return True

        now = time.time()
        if (now - self.typing_notice_time) > (TYPING_NOTICE_TIMEOUT / 1000):
            return True
        return False

    @property
    def should_send_read_marker(self):
        # type () -> bool
        """Check if we need to send out a read receipt."""
        if not self.read_markers_enabled:
            return False

        if not self.last_read_event:
            return True

        if self.last_read_event == self.last_event_id:
            return False

        return True

    @property
    def last_event_id(self):
        # type () -> str
        """Get the event id of the last shown matrix event."""
        for line in self.weechat_buffer.lines:
            for tag in line.tags:
                if tag.startswith("matrix_id"):
                    event_id = tag[10:]
                    return event_id

        return ""

    @property
    def printed_event_ids(self):
        for line in self.weechat_buffer.lines:
            for tag in line.tags:
                if tag.startswith("matrix_id"):
                    event_id = tag[10:]
                    yield event_id

    @property
    def read_markers_enabled(self):
        # type: () -> bool
        """Check if read receipts are enabled for this room."""
        return bool(int(W.string_eval_expression(
            G.CONFIG.network.read_markers_conditions,
            {},
            {"markers_enabled": str(int(self._read_markers_enabled))},
            {"type": "condition"}
        )))

    @read_markers_enabled.setter
    def read_markers_enabled(self, value):
        self._read_markers_enabled = value

    def find_nick(self, user_id):
        # type: (str) -> str
        """Find a suitable nick from a user_id."""
        if user_id in self.displayed_nicks:
            return self.displayed_nicks[user_id]

        return user_id

    def add_user(self, user_id, date, is_state, force_add=False):
        # User is already added don't add him again.
        if user_id in self.displayed_nicks:
            return

        try:
            user = self.room.users[user_id]
        except KeyError:
            # No user found, he must have left already in an event that is
            # yet to come, so do nothing
            return

        # Adding users to the nicklist is a O(1) + search time
        # operation (the nicks are added to a linked list sorted).
        # The search time is O(N * min(a,b)) where N is the number
        # of nicks already added and a/b are the length of
        # the strings that are compared at every iteration.
        # Because the search time get's increasingly longer we're
        # going to stop adding inactive users, they will be lazily added if
        # they become active.
        if is_state and not force_add and user.power_level <= 0:
            if (len(self.displayed_nicks) >=
                    G.CONFIG.network.max_nicklist_users):
                self.inactive_users.append(user_id)
                return

        try:
            self.inactive_users.remove(user_id)
        except ValueError:
            pass

        short_name = shorten_sender(user.user_id)

        # TODO handle this special case for discord bridge users and
        # freenode bridge users better
        if (user.user_id.startswith("@_discord_") or
                user.user_id.startswith("@_slack_") or
                user.user_id.startswith("@_discordpuppet_") or
                user.user_id.startswith("@_slackpuppet_") or
                user.user_id.startswith("@whatsapp_") or
                user.user_id.startswith("@facebook_") or
                user.user_id.startswith("@telegram_") or
                user.user_id.startswith("@_telegram_") or
                user.user_id.startswith("@_xmpp_") or
                user.user_id.startswith("@irc_")):
            if user.display_name:
                short_name = user.display_name[0:50]
        elif user.user_id.startswith("@twilio_"):
            short_name = shorten_sender(user.user_id[7:])
        elif user.user_id.startswith("@freenode_"):
            short_name = shorten_sender(user.user_id[9:])
        elif user.user_id.startswith("@_ircnet_"):
            short_name = shorten_sender(user.user_id[8:])
        elif user.user_id.startswith("@_oftc_"):
            short_name = shorten_sender(user.user_id[6:])
        elif user.user_id.startswith("@gitter_"):
            short_name = shorten_sender(user.user_id[7:])

        # TODO make this configurable
        if not short_name or short_name in self.displayed_nicks.values():
            # Use the full user id, but don't include the @
            nick = user_id[1:]
        else:
            nick = short_name

        buffer_user = RoomUser(nick, user_id, user.power_level, date)
        self.displayed_nicks[user_id] = nick

        if self.room.own_user_id == user_id:
            buffer_user.color = "weechat.color.chat_nick_self"
            user.nick_color = "weechat.color.chat_nick_self"

        self.weechat_buffer.join(buffer_user, date, not is_state)

    def handle_membership_events(self, event, is_state):
        date = server_ts_to_weechat(event.server_timestamp)

        if event.content["membership"] == "join":
            if (event.state_key not in self.displayed_nicks
                    and event.state_key not in self.inactive_users):
                if len(self.room.users) > 100:
                    self.unhandled_users.append(event.state_key)
                    return

                self.add_user(event.state_key, date, is_state)
            else:
                # TODO print out profile changes
                return

        elif event.content["membership"] == "leave":
            if event.state_key in self.unhandled_users:
                self.unhandled_users.remove(event.state_key)
                return

            nick = self.find_nick(event.state_key)
            if event.sender == event.state_key:
                self.weechat_buffer.part(nick, date, not is_state)
            else:
                self.weechat_buffer.kick(nick, date, not is_state)

            if event.state_key in self.displayed_nicks:
                del self.displayed_nicks[event.state_key]

            # We left the room, remember the event id of our leave, if we
            # rejoin we get events that came before this event as well as
            # after our leave, this way we know where to continue
            if event.state_key == self.room.own_user_id:
                self.leave_event_id = event.event_id

        elif event.content["membership"] == "invite":
            if is_state:
                return

            self.weechat_buffer.invite(event.state_key, date)
            return

        self.update_buffer_name()

    def update_buffer_name(self):
        if self.room.is_named:
            if self.room.name and self.room.name != "#":
                room_name = self.room.name
                room_name = (room_name if room_name.startswith("#")
                             else "#" + room_name)
            elif self.room.canonical_alias:
                room_name = self.room.canonical_alias
                self.update_canonical_alias_localvar()
            elif self.room.name == "#":
                room_name = "##"
        else:
            room_name = self.room.display_name

        if room_name is None:
            # Use placeholder room name
            room_name = 'Empty room (?)'

        self.weechat_buffer.short_name = room_name

        if G.CONFIG.human_buffer_names:
            buffer_name = "{}.{}".format(self.server_name, room_name)
            self.weechat_buffer.name = buffer_name

    def update_canonical_alias_localvar(self):
        W.buffer_set(
            self.weechat_buffer._ptr,
            "localvar_set_canonical_alias",
            self.room.canonical_alias
        )

    def _redact_line(self, event):
        def predicate(event_id, line):
            def already_redacted(tags):
                if SCRIPT_NAME + "_redacted" in tags:
                    return True
                return False

            event_tag = SCRIPT_NAME + "_id_{}".format(event_id)
            tags = line.tags

            if event_tag in tags and not already_redacted(tags):
                return True

            return False

        def redact_string(message):
            new_message = ""

            if G.CONFIG.look.redactions == RedactType.STRIKETHROUGH:
                plaintext_msg = W.string_remove_color(message, "")
                new_message = string_strikethrough(plaintext_msg)
            elif G.CONFIG.look.redactions == RedactType.NOTICE:
                new_message = message
            elif G.CONFIG.look.redactions == RedactType.DELETE:
                pass

            return new_message

        lines = self.weechat_buffer.find_lines(
            partial(predicate, event.redacts)
        )

        # No line to redact, return early
        if not lines:
            return

        censor = self.find_nick(event.sender)
        redaction_msg = Render.redacted(censor, event.reason)

        line = lines[0]
        message = line.message
        tags = line.tags

        new_message = redact_string(message)
        message = " ".join(s for s in [new_message, redaction_msg] if s)
        tags.append("matrix_redacted")

        line.message = message
        line.tags = tags

        for line in lines[1:]:
            message = line.message
            tags = line.tags

            new_message = redact_string(message)

            if not new_message:
                new_message = redaction_msg
            elif G.CONFIG.look.redactions == RedactType.NOTICE:
                new_message += " {}".format(redaction_msg)

            tags.append("matrix_redacted")

            line.message = new_message
            line.tags = tags

    def _handle_topic(self, event, is_state):
        nick = self.find_nick(event.sender)

        self.weechat_buffer.change_topic(
            nick,
            event.topic,
            server_ts_to_weechat(event.server_timestamp),
            not is_state,
        )

    @staticmethod
    def get_event_tags(event):
        # type: (Event) -> List[str]
        tags = [SCRIPT_NAME + "_id_{}".format(event.event_id)]
        if event.sender_key:
            tags.append(SCRIPT_NAME + "_senderkey_{}".format(event.sender_key))
        if event.session_id:
            tags.append(SCRIPT_NAME + "_session_id_{}".format(
                event.session_id
            ))

        return tags

    def _handle_power_level(self, _):
        for user_id in self.room.power_levels.users:
            if user_id in self.displayed_nicks:
                nick = self.find_nick(user_id)

                user = self.weechat_buffer.users[nick]
                user.power_level = self.room.power_levels.get_user_level(
                    user_id
                )

                # There is no way to change the group of a user without
                # removing him from the nicklist
                self.weechat_buffer.remove_user_from_nicklist(user)
                self.weechat_buffer._add_user_to_nicklist(user)

    def handle_state_event(self, event):
        if isinstance(event, RoomMemberEvent):
            self.handle_membership_events(event, True)
        elif isinstance(event, RoomTopicEvent):
            self._handle_topic(event, True)
        elif isinstance(event, PowerLevelsEvent):
            self._handle_power_level(event)
        elif isinstance(event, (RoomNameEvent, RoomAliasEvent)):
            self.update_buffer_name()
        elif isinstance(event, RoomEncryptionEvent):
            pass

    def handle_own_message_in_timeline(self, event):
        """Check if our own message is already printed if not print it.
        This function is called for messages that contain a transaction id
        indicating that they were sent out using our own client. If we sent out
        a message but never got a valid server response (e.g. due to
        disconnects) this function prints them out using data from the next
        sync response"""
        uuid = UUID(event.transaction_id)
        message = self.sent_messages_queue.pop(uuid, None)

        # We already got a response to the room_send_message() API call and
        # handled the message, no need to print it out again
        if not message:
            return

        message.event_id = event.event_id
        if uuid in self.printed_before_ack_queue:
            self.replace_printed_line_by_uuid(
                event.transaction_id,
                message
            )
            self.printed_before_ack_queue.remove(uuid)
            return

        if isinstance(message, OwnAction):
            self.self_action(message)
        elif isinstance(message, OwnMessage):
            self.self_message(message)
        return

    def print_room_message(self, event, extra_tags=None):
        extra_tags = extra_tags or []
        nick = self.find_nick(event.sender)

        data = Render.message(event.body, event.formatted_body)

        extra_prefix = (self.warning_prefix if event.decrypted
                        and not event.verified else "")

        date = server_ts_to_weechat(event.server_timestamp)
        self.weechat_buffer.message(
            nick, data, date, self.get_event_tags(event) + extra_tags,
            extra_prefix
        )

    def print_room_emote(self, event, extra_tags=None):
        extra_tags = extra_tags or []
        nick = self.find_nick(event.sender)
        date = server_ts_to_weechat(event.server_timestamp)

        extra_prefix = (self.warning_prefix if event.decrypted
                        and not event.verified else "")

        self.weechat_buffer.action(
            nick, event.body, date, self.get_event_tags(event) + extra_tags,
            extra_prefix
        )

    def print_room_notice(self, event, extra_tags=None):
        extra_tags = extra_tags or []
        nick = self.find_nick(event.sender)
        date = server_ts_to_weechat(event.server_timestamp)
        extra_prefix = (self.warning_prefix if event.decrypted
                        and not event.verified else "")

        self.weechat_buffer.notice(
            nick, event.body, date, self.get_event_tags(event) + extra_tags,
            extra_prefix
        )

    def print_room_media(self, event, extra_tags=None):
        extra_tags = extra_tags or []
        nick = self.find_nick(event.sender)
        date = server_ts_to_weechat(event.server_timestamp)
        if isinstance(event, RoomMessageMedia):
            data = Render.media(event.url, event.body, self.homeserver.geturl())
        else:
            data = Render.encrypted_media(
                event.url, event.body, event.key["k"], event.hashes["sha256"],
                event.iv, self.homeserver.geturl()
            )

        extra_prefix = (self.warning_prefix if event.decrypted
                        and not event.verified else "")

        self.weechat_buffer.message(
            nick, data, date, self.get_event_tags(event) + extra_tags,
            extra_prefix
        )

    def print_unknown(self, event, extra_tags=None):
        extra_tags = extra_tags or []
        nick = self.find_nick(event.sender)
        date = server_ts_to_weechat(event.server_timestamp)
        data = Render.unknown(event.type, event.content)
        extra_prefix = (self.warning_prefix if event.decrypted
                        and not event.verified else "")

        self.weechat_buffer.message(
            nick, data, date, self.get_event_tags(event) + extra_tags,
            extra_prefix
        )

    def print_redacted(self, event, extra_tags=None):
        extra_tags = extra_tags or []

        nick = self.find_nick(event.sender)
        date = server_ts_to_weechat(event.server_timestamp)
        tags = self.get_event_tags(event)
        tags.append(SCRIPT_NAME + "_redacted")
        tags += extra_tags

        censor = self.find_nick(event.redacter)
        data = Render.redacted(censor, event.reason)

        self.weechat_buffer.message(nick, data, date, tags)

    def print_room_encryption(self, event, extra_tags=None):
        nick = self.find_nick(event.sender)
        data = Render.room_encryption(nick)
        # TODO this should also have tags
        self.weechat_buffer.info(data)

    def print_megolm(self, event, extra_tags=None):
        extra_tags = extra_tags or []
        nick = self.find_nick(event.sender)
        date = server_ts_to_weechat(event.server_timestamp)

        data = Render.megolm()

        session_id_tag = SCRIPT_NAME + "_sessionid_" + event.session_id
        self.weechat_buffer.message(
            nick,
            data,
            date,
            self.get_event_tags(event) + [session_id_tag] + extra_tags
        )

        self.undecrypted_events.append(event)

    def print_bad_event(self, event, extra_tags=None):
        extra_tags = extra_tags or []
        nick = self.find_nick(event.sender)
        date = server_ts_to_weechat(event.server_timestamp)
        data = Render.bad(event)
        extra_prefix = self.warning_prefix

        self.weechat_buffer.message(
            nick, data, date, self.get_event_tags(event) + extra_tags,
            extra_prefix
        )

    def handle_room_messages(self, event, extra_tags=None):
        if isinstance(event, RoomMessageEmote):
            self.print_room_emote(event, extra_tags)

        elif isinstance(event, RoomMessageText):
            self.print_room_message(event, extra_tags)

        elif isinstance(event, RoomMessageNotice):
            self.print_room_notice(event, extra_tags)

        elif isinstance(event, RoomMessageMedia):
            self.print_room_media(event, extra_tags)

        elif isinstance(event, RoomEncryptedMedia):
            self.print_room_media(event, extra_tags)

        elif isinstance(event, RoomMessageUnknown):
            self.print_unknown(event, extra_tags)

        elif isinstance(event, RoomEncryptionEvent):
            self.print_room_encryption(event, extra_tags)

        elif isinstance(event, MegolmEvent):
            self.print_megolm(event, extra_tags)

    def force_load_member(self, event):
        if (event.sender not in self.displayed_nicks and
            event.sender in self.room.users):

            try:
                self.unhandled_users.remove(event.sender)
            except ValueError:
                pass

            self.add_user(event.sender, 0, True, True)

    def handle_timeline_event(self, event, extra_tags=None):
        # TODO this should be done for every messagetype that gets printed in
        # the buffer
        if isinstance(event, (RoomMessage, MegolmEvent)):
            self.force_load_member(event)

        if event.transaction_id:
            self.handle_own_message_in_timeline(event)
            return

        if isinstance(event, RoomMemberEvent):
            self.handle_membership_events(event, False)

        elif isinstance(event, (RoomNameEvent, RoomAliasEvent)):
            self.update_buffer_name()

        elif isinstance(event, RoomTopicEvent):
            self._handle_topic(event, False)

        # Emotes are a subclass of RoomMessageText, so put them before the text
        # ones
        elif isinstance(event, RoomMessageEmote):
            self.print_room_emote(event, extra_tags)

        elif isinstance(event, RoomMessageText):
            self.print_room_message(event, extra_tags)

        elif isinstance(event, RoomMessageNotice):
            self.print_room_notice(event, extra_tags)

        elif isinstance(event, RoomMessageMedia):
            self.print_room_media(event, extra_tags)

        elif isinstance(event, RoomEncryptedMedia):
            self.print_room_media(event, extra_tags)

        elif isinstance(event, RoomMessageUnknown):
            self.print_unknown(event, extra_tags)

        elif isinstance(event, RedactionEvent):
            self._redact_line(event)

        elif isinstance(event, RedactedEvent):
            self.print_redacted(event, extra_tags)

        elif isinstance(event, RoomEncryptionEvent):
            self.print_room_encryption(event, extra_tags)

        elif isinstance(event, PowerLevelsEvent):
            # TODO we should print out a message for this event
            self._handle_power_level(event)

        elif isinstance(event, MegolmEvent):
            self.print_megolm(event, extra_tags)

        elif isinstance(event, UnknownEvent):
            pass

        elif isinstance(event, BadEvent):
            self.print_bad_event(event, extra_tags)

        elif isinstance(event, UnknownBadEvent):
            self.error("Unknown bad event: {}".format(
                pprint.pformat(event.source)
            ))

        else:
            W.prnt(
                "", "Unhandled event of type {}.".format(type(event).__name__)
            )

    def self_message(self, message):
        # type: (OwnMessage) -> None
        nick = self.find_nick(self.room.own_user_id)
        data = message.formatted_message.to_weechat()
        if message.event_id:
            tags = [SCRIPT_NAME + "_id_{}".format(message.event_id)]
        else:
            tags = [SCRIPT_NAME + "_uuid_{}".format(message.uuid)]
        date = message.age

        self.weechat_buffer.self_message(nick, data, date, tags)

    def self_action(self, message):
        # type: (OwnMessage) -> None
        nick = self.find_nick(self.room.own_user_id)
        date = message.age
        if message.event_id:
            tags = [SCRIPT_NAME + "_id_{}".format(message.event_id)]
        else:
            tags = [SCRIPT_NAME + "_uuid_{}".format(message.uuid)]

        self.weechat_buffer.self_action(
            nick, message.formatted_message.to_weechat(), date, tags
        )

    @staticmethod
    def _find_by_uuid_predicate(uuid, line):
        uuid_tag = SCRIPT_NAME + "_uuid_{}".format(uuid)
        tags = line.tags

        if uuid_tag in tags:
            return True
        return False

    def mark_message_as_unsent(self, uuid, _):
        """Append to already printed lines that are greyed out an error
        message"""
        lines = self.weechat_buffer.find_lines(
            partial(self._find_by_uuid_predicate, uuid)
        )
        last_line = lines[-1]

        message = last_line.message
        message += (" {del_color}<{ncolor}{error_color}Error sending "
                    "message{del_color}>{ncolor}").format(
            del_color=W.color("chat_delimiters"),
            ncolor=W.color("reset"),
            error_color=W.color(color_pair(
                G.CONFIG.color.error_message_fg,
                G.CONFIG.color.error_message_bg)))

        last_line.message = message

    def replace_printed_line_by_uuid(self, uuid, new_message):
        """Replace already printed lines that are greyed out with real ones."""
        if isinstance(new_message, OwnAction):
            displayed_nick = self.displayed_nicks[self.room.own_user_id]
            user = self.weechat_buffer._get_user(displayed_nick)
            data = self.weechat_buffer._format_action(
                user,
                new_message.formatted_message.to_weechat()
            )
            new_lines = data.split("\n")
        else:
            new_lines = new_message.formatted_message.to_weechat().split("\n")

        line_count = len(new_lines)

        lines = self.weechat_buffer.find_lines(
            partial(self._find_by_uuid_predicate, uuid), line_count
        )

        for i, line in enumerate(lines):
            line.message = new_lines[i]
            tags = line.tags

            new_tags = [
                tag for tag in tags
                if not tag.startswith(SCRIPT_NAME + "_uuid_")
            ]
            new_tags.append(SCRIPT_NAME + "_id_" + new_message.event_id)
            line.tags = new_tags

    def replace_undecrypted_line(self, event):
        """Find an undecrypted message in the buffer and replace it with the now
        decrypted event."""
        # TODO different messages need different formatting
        # To implement this, refactor out the different formatting code
        # snippets to a Formatter class and reuse them here.
        if not isinstance(event, RoomMessageText):
            return

        def predicate(event_id, line):
            event_tag = SCRIPT_NAME + "_id_{}".format(event_id)
            if event_tag in line.tags:
                return True
            return False

        lines = self.weechat_buffer.find_lines(
            partial(predicate, event.event_id)
        )

        if not lines:
            return

        formatted = None
        if event.formatted_body:
            formatted = Formatted.from_html(event.formatted_body)

        data = formatted.to_weechat() if formatted else event.body
        # TODO this isn't right if the data has multiple lines, that is
        # everything is printed on a single line and newlines are shown as a
        # space.
        # Weechat should support deleting lines and printing new ones at an
        # arbitrary position.
        # To implement this without weechat support either only handle single
        # line messages or edit the first line in place, print new ones at the
        # bottom and sort the buffer lines.
        lines[0].message = data

    def old_message(self, event):
        tags = list(self.weechat_buffer.tags["old_message"])
        # TODO events that change the room state (topics, membership changes,
        # etc...) should be printed out as well, but for this to work without
        # messing up the room state the state change will need to be separated
        # from the print logic.
        if isinstance(event, RoomMessage):
            self.force_load_member(event)
            self.handle_room_messages(event, tags)

        elif isinstance(event, MegolmEvent):
            self.print_megolm(event, tags)

        elif isinstance(event, RedactedEvent):
            self.print_redacted(event, tags)

        elif isinstance(event, BadEvent):
            self.print_bad_event(event, tags)

    def sort_messages(self):
        class LineCopy(object):
            def __init__(
                self, date, date_printed, tags, prefix, message, highlight
            ):
                self.date = date
                self.date_printed = date_printed
                self.tags = tags
                self.prefix = prefix
                self.message = message
                self.highlight = highlight

            @classmethod
            def from_line(cls, line):
                return cls(
                    line.date,
                    line.date_printed,
                    line.tags,
                    line.prefix,
                    line.message,
                    line.highlight,
                )

        lines = [
            LineCopy.from_line(line) for line in self.weechat_buffer.lines
        ]
        sorted_lines = sorted(lines, key=lambda line: line.date, reverse=True)

        for line_number, line in enumerate(self.weechat_buffer.lines):
            new = sorted_lines[line_number]
            line.update(
                new.date, new.date_printed, new.tags, new.prefix, new.message
            )

    def handle_backlog(self, response):
        self.prev_batch = response.end

        for event in response.chunk:
            # The first backlog request seems to have a race condition going on
            # where we receive a message in a sync response, get a prev_batch,
            # yet when we request older messages with the prev_batch the same
            # message might appear in the room messages response. This only
            # seems to happen if the message is relatively recently sent.
            # Because of this we check if our first backlog request contains
            # some already printed events, if so; skip printing them.
            if (self.first_backlog_request
                    and event.event_id in self.printed_event_ids):
                continue

            self.old_message(event)

        self.sort_messages()

        self.first_backlog_request = False
        self.backlog_pending = False

    def handle_joined_room(self, info):
        for event in info.state:
            self.handle_state_event(event)

        timeline_events = None

        # This is a rejoin, skip already handled events
        if not self.joined:
            leave_index = None

            for i, event in enumerate(info.timeline.events):
                if event.event_id == self.leave_event_id:
                    leave_index = i
                    break

            if leave_index:
                timeline_events = info.timeline.events[leave_index + 1:]
                # Handle our leave as a state event since we're not in the
                # nicklist anymore but we're already printed out our leave
                self.handle_state_event(info.timeline.events[leave_index])
            else:
                timeline_events = info.timeline.events

            # mark that we are now joined
            self.joined = True

        else:
            timeline_events = info.timeline.events

        for event in timeline_events:
            self.handle_timeline_event(event)

        for event in info.account_data:
            if isinstance(event, FullyReadEvent):
                if event.event_id == self.last_event_id:
                    current_buffer = W.buffer_search("", "")

                    if self.weechat_buffer._ptr == current_buffer:
                        continue

                    W.buffer_set(self.weechat_buffer._ptr, "unread", "")
                    W.buffer_set(self.weechat_buffer._ptr, "hotlist", "-1")

        # We didn't handle all joined users, the room display name might still
        # be outdated because of that, update it now.
        if self.unhandled_users:
            self.update_buffer_name()

    def handle_left_room(self, info):
        self.joined = False

        for event in info.state:
            self.handle_state_event(event)

        for event in info.timeline.events:
            self.handle_timeline_event(event)

    def error(self, string):
        # type: (str) -> None
        self.weechat_buffer.error(string)



# -*- coding: utf-8 -*-

# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Module implementing upload functionality."""

from __future__ import unicode_literals

import attr
import time
import json
from typing import Dict, Any
from uuid import uuid1, UUID
from enum import Enum

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError  # type: ignore

from .globals import SCRIPT_NAME, SERVERS, W, UPLOADS
from .utf import utf8_decode
from .message_renderer import Render
from matrix import globals as G
from nio import Api


class UploadState(Enum):
    created = 0
    active = 1
    finished = 2
    error = 3
    aborted = 4


@attr.s
class Proxy(object):
    ptr = attr.ib(type=str)

    @property
    def name(self):
        return W.infolist_string(self.ptr, "name")

    @property
    def address(self):
        return W.infolist_string(self.ptr, "address")

    @property
    def type(self):
        return W.infolist_string(self.ptr, "type_string")

    @property
    def port(self):
        return str(W.infolist_integer(self.ptr, "port"))

    @property
    def user(self):
        return W.infolist_string(self.ptr, "username")

    @property
    def password(self):
        return W.infolist_string(self.ptr, "password")


@attr.s
class Upload(object):
    """Class representing an upload to a matrix server."""

    server_name = attr.ib(type=str)
    server_address = attr.ib(type=str)
    access_token = attr.ib(type=str)
    room_id = attr.ib(type=str)
    filepath = attr.ib(type=str)
    encrypt = attr.ib(type=bool, default=False)
    file_keys = attr.ib(type=Dict, default=None)

    done = 0
    total = 0

    uuid = None
    buffer = None
    upload_hook = None
    content_uri = None
    file_name = None
    mimetype = "?"
    state = UploadState.created

    def __attrs_post_init__(self):
        self.uuid = uuid1()
        self.buffer = ""

        server = SERVERS[self.server_name]

        proxy_name = server.config.proxy
        proxy = None
        proxies_list = None

        if proxy_name:
            proxies_list = W.infolist_get("proxy", "", proxy_name)
            if proxies_list:
                W.infolist_next(proxies_list)
                proxy = Proxy(proxies_list)

        process_args = {
            "arg1": self.filepath,
            "arg2": self.server_address,
            "arg3": self.access_token,
            "buffer_flush": "1",
        }

        arg_count = 3

        if self.encrypt:
            arg_count += 1
            process_args["arg{}".format(arg_count)] = "--encrypt"

        if not server.config.ssl_verify:
            arg_count += 1
            process_args["arg{}".format(arg_count)] = "--insecure"

        if proxy:
            arg_count += 1
            process_args["arg{}".format(arg_count)] = "--proxy-type"
            arg_count += 1
            process_args["arg{}".format(arg_count)] = proxy.type

            arg_count += 1
            process_args["arg{}".format(arg_count)] = "--proxy-address"
            arg_count += 1
            process_args["arg{}".format(arg_count)] = proxy.address

            arg_count += 1
            process_args["arg{}".format(arg_count)] = "--proxy-port"
            arg_count += 1
            process_args["arg{}".format(arg_count)] = proxy.port

            if proxy.user:
                arg_count += 1
                process_args["arg{}".format(arg_count)] = "--proxy-user"
                arg_count += 1
                process_args["arg{}".format(arg_count)] = proxy.user

            if proxy.password:
                arg_count += 1
                process_args["arg{}".format(arg_count)] = "--proxy-password"
                arg_count += 1
                process_args["arg{}".format(arg_count)] = proxy.password

        self.upload_hook = W.hook_process_hashtable(
            "matrix_upload",
            process_args,
            0,
            "upload_cb",
            str(self.uuid)
        )

        if proxies_list:
            W.infolist_free(proxies_list)

    def abort(self):
        pass

    @property
    def msgtype(self):
        # type: () -> str
        assert self.mimetype
        return Api.mimetype_to_msgtype(self.mimetype)

    @property
    def content(self):
        # type: () -> Dict[Any, Any]
        assert self.content_uri

        if self.encrypt:
            content = {
                "body": self.file_name,
                "msgtype": self.msgtype,
                "file": self.file_keys,
            }
            content["file"]["url"] = self.content_uri
            content["file"]["mimetype"] = self.mimetype

            # TODO thumbnail if it's an image

            return content

        return {
            "msgtype": self.msgtype,
            "body": self.file_name,
            "url": self.content_uri,
        }

    @property
    def render(self):
        # type: () -> str
        assert self.content_uri

        if self.encrypt:
            return Render.encrypted_media(
                self.content_uri,
                self.file_name,
                self.file_keys["key"]["k"],
                self.file_keys["hashes"]["sha256"],
                self.file_keys["iv"],
            )

        return Render.media(self.content_uri, self.file_name)


@attr.s
class UploadsBuffer(object):
    """Weechat buffer showing the uploads for a server."""

    _ptr = ""           # type: str
    _selected_line = 0  # type: int
    uploads = UPLOADS

    def __attrs_post_init__(self):
        self._ptr = W.buffer_new(
            SCRIPT_NAME + ".uploads",
            "",
            "",
            "",
            "",
        )
        W.buffer_set(self._ptr, "type", "free")
        W.buffer_set(self._ptr, "title", "Upload list")
        W.buffer_set(self._ptr, "key_bind_meta2-A", "/uploads up")
        W.buffer_set(self._ptr, "key_bind_meta2-B", "/uploads down")
        W.buffer_set(self._ptr, "localvar_set_type", "uploads")

        self.render()

    def move_line_up(self):
        self._selected_line = max(self._selected_line - 1, 0)
        self.render()

    def move_line_down(self):
        self._selected_line = min(
            self._selected_line + 1,
            len(self.uploads) - 1
        )
        self.render()

    def display(self):
        """Display the buffer."""
        W.buffer_set(self._ptr, "display", "1")

    def render(self):
        """Render the new state of the upload buffer."""
        # This function is under the MIT license.
        # Copyright (c) 2016 Vladimir Ignatev
        def progress(count, total):
            bar_len = 60

            if total == 0:
                bar = '-' * bar_len
                return "[{}] {}%".format(bar, "?")

            filled_len = int(round(bar_len * count / float(total)))
            percents = round(100.0 * count / float(total), 1)
            bar = '=' * filled_len + '-' * (bar_len - filled_len)

            return "[{}] {}%".format(bar, percents)

        W.buffer_clear(self._ptr)
        header = "{}{}{}{}{}{}{}{}".format(
            W.color("green"),
            "Actions (letter+enter):",
            W.color("lightgreen"),
            "  [A] Accept",
            "  [C] Cancel",
            "  [R] Remove",
            "  [P] Purge finished",
            "  [Q] Close this buffer"
        )
        W.prnt_y(self._ptr, 0, header)

        for line_number, upload in enumerate(self.uploads.values()):
            line_color = "{},{}".format(
                "white" if line_number == self._selected_line else "default",
                "blue" if line_number == self._selected_line else "default",
            )
            first_line = ("%s%s %-24s %s%s%s %s (%s.%s)" % (
                          W.color(line_color),
                          "*** " if line_number == self._selected_line else "    ",
                          upload.room_id,
                          "\"",
                          upload.filepath,
                          "\"",
                          upload.mimetype,
                          SCRIPT_NAME,
                          upload.server_name,
                          ))
            W.prnt_y(self._ptr, (line_number * 2) + 2, first_line)

            status_color = "{},{}".format("green", "blue")
            status = "{}{}{}".format(
                W.color(status_color),
                upload.state.name,
                W.color(line_color)
            )

            second_line = ("{color}{prefix} {status} {progressbar} "
                           "{done} / {total}").format(
                color=W.color(line_color),
                prefix="*** " if line_number == self._selected_line else "    ",
                status=status,
                progressbar=progress(upload.done, upload.total),
                done=W.string_format_size(upload.done),
                total=W.string_format_size(upload.total))

            W.prnt_y(self._ptr, (line_number * 2) + 3, second_line)


def find_upload(uuid):
    return UPLOADS.get(uuid, None)


def handle_child_message(upload, message):
    if message["type"] == "progress":
        upload.done = message["data"]

    elif message["type"] == "status":
        if message["status"] == "started":
            upload.state = UploadState.active
            upload.total = message["total"]
            upload.mimetype = message["mimetype"]
            upload.file_name = message["file_name"]

        elif message["status"] == "done":
            upload.state = UploadState.finished
            upload.content_uri = message["url"]
            upload.file_keys = message.get("file_keys", None)

            server = SERVERS.get(upload.server_name, None)

            if not server:
                return

            server.room_send_upload(upload)

        elif message["status"] == "error":
            upload.state = UploadState.error

    if G.CONFIG.upload_buffer:
        G.CONFIG.upload_buffer.render()


@utf8_decode
def upload_cb(data, command, return_code, out, err):
    upload = find_upload(UUID(data))

    if not upload:
        return W.WEECHAT_RC_OK

    if return_code == W.WEECHAT_HOOK_PROCESS_ERROR:
        W.prnt("", "Error with command '%s'" % command)
        return W.WEECHAT_RC_OK

    if err != "":
        W.prnt("", "Error with command '%s'" % err)
        upload.state = UploadState.error

    if out != "":
        upload.buffer += out
        messages = upload.buffer.split("\n")
        upload.buffer = ""

        for m in messages:
            try:
                message = json.loads(m)
            except (JSONDecodeError, TypeError):
                upload.buffer += m
                continue

            handle_child_message(upload, message)

    return W.WEECHAT_RC_OK



# -*- coding: utf-8 -*-

# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
# Copyright © 2018, 2019 Denis Kasak <dkasak@termina.org.uk>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import unicode_literals
import argparse
import os
import re
import shlex
from builtins import str
from future.moves.itertools import zip_longest
from collections import defaultdict
from functools import partial
from nio import EncryptionError, LocalProtocolError

from . import globals as G
from .colors import Formatted
from .globals import SERVERS, W, UPLOADS, SCRIPT_NAME
from .server import MatrixServer
from .utf import utf8_decode
from .utils import key_from_value, parse_redact_args
from .uploads import UploadsBuffer, Upload

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse  # type: ignore


class ParseError(Exception):
    pass


class WeechatArgParse(argparse.ArgumentParser):
    def print_usage(self, file=None):
        pass

    def error(self, message):
        message = (
            "{prefix}Error: {message} for command {command} "
            "(see /help {command})"
        ).format(prefix=W.prefix("error"), message=message, command=self.prog)
        W.prnt("", message)
        raise ParseError


class WeechatCommandParser(object):
    @staticmethod
    def _run_parser(parser, args):
        try:
            parsed_args = parser.parse_args(shlex.split(args))
            return parsed_args
        except ParseError:
            return None

    @staticmethod
    def topic(args):
        parser = WeechatArgParse(prog="topic")

        parser.add_argument("-delete", action="store_true")
        parser.add_argument("topic", nargs="*")

        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def kick(args):
        parser = WeechatArgParse(prog="kick")
        parser.add_argument("user_id")
        parser.add_argument("reason", nargs="*")

        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def invite(args):
        parser = WeechatArgParse(prog="invite")
        parser.add_argument("user_id")

        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def join(args):
        parser = WeechatArgParse(prog="join")
        parser.add_argument("room_id")
        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def part(args):
        parser = WeechatArgParse(prog="part")
        parser.add_argument("room_id", nargs="?")
        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def devices(args):
        parser = WeechatArgParse(prog="devices")
        subparsers = parser.add_subparsers(dest="subcommand")
        subparsers.add_parser("list")

        delete_parser = subparsers.add_parser("delete")
        delete_parser.add_argument("device_id")

        name_parser = subparsers.add_parser("set-name")
        name_parser.add_argument("device_id")
        name_parser.add_argument("device_name", nargs="*")

        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def olm(args):
        parser = WeechatArgParse(prog="olm")
        subparsers = parser.add_subparsers(dest="subcommand")

        info_parser = subparsers.add_parser("info")
        info_parser.add_argument(
            "category", nargs="?", default="private",
            choices=[
                "all",
                "blacklisted",
                "private",
                "unverified",
                "verified",
                "ignored"
            ])
        info_parser.add_argument("filter", nargs="?")

        verify_parser = subparsers.add_parser("verify")
        verify_parser.add_argument("user_filter")
        verify_parser.add_argument("device_filter", nargs="?")

        unverify_parser = subparsers.add_parser("unverify")
        unverify_parser.add_argument("user_filter")
        unverify_parser.add_argument("device_filter", nargs="?")

        blacklist_parser = subparsers.add_parser("blacklist")
        blacklist_parser.add_argument("user_filter")
        blacklist_parser.add_argument("device_filter", nargs="?")

        unblacklist_parser = subparsers.add_parser("unblacklist")
        unblacklist_parser.add_argument("user_filter")
        unblacklist_parser.add_argument("device_filter", nargs="?")

        ignore_parser = subparsers.add_parser("ignore")
        ignore_parser.add_argument("user_filter")
        ignore_parser.add_argument("device_filter", nargs="?")

        unignore_parser = subparsers.add_parser("unignore")
        unignore_parser.add_argument("user_filter")
        unignore_parser.add_argument("device_filter", nargs="?")

        export_parser = subparsers.add_parser("export")
        export_parser.add_argument("file")
        export_parser.add_argument("passphrase")

        import_parser = subparsers.add_parser("import")
        import_parser.add_argument("file")
        import_parser.add_argument("passphrase")

        sas_parser = subparsers.add_parser("verification")
        sas_parser.add_argument(
            "action",
            choices=[
                "start",
                "accept",
                "confirm",
                "cancel",
            ])
        sas_parser.add_argument("user_id")
        sas_parser.add_argument("device_id")

        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def room(args):
        parser = WeechatArgParse(prog="room")
        subparsers = parser.add_subparsers(dest="subcommand")
        typing_notification = subparsers.add_parser("typing-notifications")
        typing_notification.add_argument(
            "state",
            choices=["enable", "disable", "toggle"]
        )

        read_markers = subparsers.add_parser("read-markers")
        read_markers.add_argument(
            "state",
            choices=["enable", "disable", "toggle"]
        )

        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def uploads(args):
        parser = WeechatArgParse(prog="uploads")
        subparsers = parser.add_subparsers(dest="subcommand")
        subparsers.add_parser("list")
        subparsers.add_parser("listfull")
        subparsers.add_parser("up")
        subparsers.add_parser("down")

        return WeechatCommandParser._run_parser(parser, args)

    @staticmethod
    def upload(args):
        parser = WeechatArgParse(prog="upload")
        parser.add_argument("file")
        return WeechatCommandParser._run_parser(parser, args)


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def partition_key(key):
    groups = grouper(key, 4, " ")
    return ' '.join(''.join(g) for g in groups)


def hook_commands():
    W.hook_command(
        # Command name and short description
        "matrix",
        "Matrix chat protocol command",
        # Synopsis
        (
            "server add <server-name> <hostname>[:<port>] ||"
            "server delete|list|listfull <server-name> ||"
            "connect <server-name> ||"
            "disconnect <server-name> ||"
            "reconnect <server-name> ||"
            "help <matrix-command>"
        ),
        # Description
        (
            "    server: list, add, or remove Matrix servers\n"
            "   connect: connect to Matrix servers\n"
            "disconnect: disconnect from one or all Matrix servers\n"
            " reconnect: reconnect to server(s)\n"
            "      help: show detailed command help\n\n"
            "Use /matrix help [command] to find out more.\n"
        ),
        # Completions
        (
            "server %(matrix_server_commands)|%* ||"
            "connect %(matrix_servers) ||"
            "disconnect %(matrix_servers) ||"
            "reconnect %(matrix_servers) ||"
            "help %(matrix_commands)"
        ),
        # Function name
        "matrix_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "redact",
        "redact messages",
        # Synopsis
        ('<event-id>[:"<message-part>"] [<reason>]'),
        # Description
        (
            "    event-id: event id of the message that will be redacted\n"
            "message-part: an initial part of the message (ignored, only "
            "used\n"
            "              as visual feedback when using completion)\n"
            "      reason: the redaction reason\n"
        ),
        # Completions
        ("%(matrix_messages)"),
        # Function name
        "matrix_redact_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "reply-matrix",
        "reply to a message",
        # Synopsis
        ('<event-id>[:"<message-part>"] [<reply>]'),
        # Description
        (
            "    event-id: event id of the message that will be replied to\n"
            "message-part: an initial part of the message (ignored, only "
            "used\n"
            "              as visual feedback when using completion)\n"
            "       reply: the reply\n"
        ),
        # Completions
        ("%(matrix_messages)"),
        # Function name
        "matrix_reply_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "topic",
        "get/set the room topic",
        # Synopsis
        ("[<topic>|-delete]"),
        # Description
        ("  topic: topic to set\n" "-delete: delete room topic"),
        # Completions
        "",
        # Callback
        "matrix_topic_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "me",
        "send an emote message to the current room",
        # Synopsis
        ("<message>"),
        # Description
        ("message: message to send"),
        # Completions
        "",
        # Callback
        "matrix_me_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "kick",
        "kick a user from the current room",
        # Synopsis
        ("<user-id> [<reason>]"),
        # Description
        (
            "user-id: user-id to kick\n"
            " reason: reason why the user was kicked"
        ),
        # Completions
        ("%(matrix_users)"),
        # Callback
        "matrix_kick_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "invite",
        "invite a user to the current room",
        # Synopsis
        ("<user-id>"),
        # Description
        ("user-id: user-id to invite"),
        # Completions
        ("%(matrix_users)"),
        # Callback
        "matrix_invite_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "join",
        "join a room",
        # Synopsis
        ("<room-id>|<room-alias>"),
        # Description
        (
            "   room-id: room-id of the room to join\n"
            "room-alias: room alias of the room to join"
        ),
        # Completions
        "",
        # Callback
        "matrix_join_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "part",
        "leave a room",
        # Synopsis
        ("[<room-name>]"),
        # Description
        ("   room-name: room name of the room to leave"),
        # Completions
        "",
        # Callback
        "matrix_part_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "devices",
        "list, delete or rename matrix devices",
        # Synopsis
        ("list ||"
         "delete <device-id> ||"
         "set-name <name>"
         ),
        # Description
        ("device-id: device id of the device to delete\n"
         "     name: new device name to set\n"),
        # Completions
        ("list ||"
         "delete %(matrix_own_devices) ||"
         "set-name %(matrix_own_devices)"),
        # Callback
        "matrix_devices_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "olm",
        "Matrix olm encryption configuration command",
        # Synopsis
        ("info all|blacklisted|ignored|private|unverified|verified <filter>||"
         "blacklist <user-id> <device-id> ||"
         "unverify <user-id> <device-id> ||"
         "verify <user-id> <device-id> ||"
         "verification start|accept|cancel|confirm <user-id> <device-id> ||"
         "ignore <user-id> <device-id> ||"
         "unignore <user-id> <device-id> ||"
         "export <file-name> <passphrase> ||"
         "import <file-name> <passphrase>"
         ),
        # Description
        ("       info: show info about known devices and their keys\n"
         "  blacklist: blacklist a device\n"
         "unblacklist: unblacklist a device\n"
         "   unverify: unverify a device\n"
         "     verify: verify a device\n"
         "     ignore: ignore an unverifiable but non-blacklist-worthy device\n"
         "   unignore: unignore a device\n"
         "verification: manage interactive device verification\n"
         "     export: export encryption keys\n"
         "     import: import encryption keys\n\n"
         "Examples:"
         "\n  /olm verify @example:example.com *"
         "\n  /olm info all example*"
         ),
        # Completions
        ('info all|blacklisted|ignored|private|unverified|verified ||'
         'blacklist %(olm_user_ids) %(olm_devices) ||'
         'unblacklist %(olm_user_ids) %(olm_devices) ||'
         'unverify %(olm_user_ids) %(olm_devices) ||'
         'verify %(olm_user_ids) %(olm_devices) ||'
         'verification start|accept|cancel|confirm %(olm_user_ids) %(olm_devices) ||'
         'ignore %(olm_user_ids) %(olm_devices) ||'
         'unignore %(olm_user_ids) %(olm_devices) ||'
         'export %(filename) ||'
         'import %(filename)'
         ),
        # Function name
        'matrix_olm_command_cb',
        '')

    W.hook_command(
        # Command name and short description
        "room",
        "change room state",
        # Synopsis
        ("typing-notifications <state>||"
         "read-markers <state>"
         ),
        # Description
        ("state: one of enable, disable or toggle\n"),
        # Completions
        ("typing-notifications enable|disable|toggle||"
         "read-markers enable|disable|toggle"
        ),
        # Callback
        "matrix_room_command_cb",
        "",
    )

    # W.hook_command(
    #     # Command name and short description
    #     "uploads",
    #     "Open the uploads buffer or list uploads in the core buffer",
    #     # Synopsis
    #     ("list||"
    #      "listfull"
    #      ),
    #     # Description
    #     (""),
    #     # Completions
    #     ("list ||"
    #      "listfull"),
    #     # Callback
    #     "matrix_uploads_command_cb",
    #     "",
    # )

    W.hook_command(
        # Command name and short description
        "upload",
        "Upload a file to a room",
        # Synopsis
        ("<file>"),
        # Description
        (""),
        # Completions
        ("%(filename)"),
        # Callback
        "matrix_upload_command_cb",
        "",
    )

    W.hook_command(
        # Command name and short description
        "send-anyways",
        "Send the last message in a room ignorin unverified devices.",
        # Synopsis
        "",
        # Description
        "Send the last message in a room despite there being unverified "
        "devices. The unverified devices will be marked as ignored after "
        "running this command.",
        # Completions
        "",
        # Callback
        "matrix_send_anyways_cb",
        "",
    )

    W.hook_command_run("/buffer clear", "matrix_command_buf_clear_cb", "")

    if G.CONFIG.network.fetch_backlog_on_pgup:
        hook_page_up()


def hook_key_bindings():
    W.hook_hsignal("matrix_cursor_reply", "matrix_cursor_reply_signal_cb", "")

    binding = "@chat(python.{}*):r".format(G.BUFFER_NAME_PREFIX)
    W.key_bind("cursor", {
        binding: "hsignal:matrix_cursor_reply",
    })


def format_device(device_id, fp_key, display_name):
    fp_key = partition_key(fp_key)
    message = ("    - Device ID:      {device_color}{device_id}{ncolor}\n"
               "      - Display name: {device_color}{display_name}{ncolor}\n"
               "      - Device key:   {key_color}{fp_key}{ncolor}").format(
                       device_color=W.color("chat_channel"),
                       device_id=device_id,
                       ncolor=W.color("reset"),
                       display_name=display_name,
                       key_color=W.color("chat_server"),
                       fp_key=fp_key)
    return message


def olm_info_command(server, args):
    def print_devices(
        device_store,
        filter_regex,
        device_category="All",
        predicate=None,
    ):
        user_strings = []
        try:
            filter_regex = re.compile(args.filter) if args.filter else None
        except re.error as e:
            server.error("Invalid regular expression: {}.".format(e.args[0]))
            return

        for user_id in sorted(device_store.users):
            device_strings = []
            for device in device_store.active_user_devices(user_id):
                if filter_regex:
                    if (not filter_regex.search(user_id) and
                            not filter_regex.search(device.id)):
                        continue

                if predicate:
                    if not predicate(device):
                        continue

                device_strings.append(format_device(
                    device.id,
                    device.ed25519,
                    device.display_name
                ))

            if not device_strings:
                continue

            d_string = "\n".join(device_strings)
            message = ("  - User: {user_color}{user}{ncolor}\n").format(
                    user_color=W.color("chat_nick"),
                    user=user_id,
                    ncolor=W.color("reset"))
            message += d_string
            user_strings.append(message)

        if not user_strings:
            message = ("{prefix}matrix: No matching devices "
                       "found.").format(prefix=W.prefix("error"))
            W.prnt(server.server_buffer, message)
            return

        server.info("{} devices:\n".format(device_category))
        W.prnt(server.server_buffer, "\n".join(user_strings))

    olm = server.client.olm

    if not hasattr(args, 'category') or args.category == "private":
        fp_key = partition_key(olm.account.identity_keys["ed25519"])
        message = ("Identity keys:\n"
                   "  - User:        {user_color}{user}{ncolor}\n"
                   "  - Device ID:   {device_color}{device_id}{ncolor}\n"
                   "  - Device key:  {key_color}{fp_key}{ncolor}\n"
                   "").format(
                       user_color=W.color("chat_self"),
                       ncolor=W.color("reset"),
                       user=olm.user_id,
                       device_color=W.color("chat_channel"),
                       device_id=olm.device_id,
                       key_color=W.color("chat_server"),
                       fp_key=fp_key)
        server.info(message)

    elif args.category == "all":
        print_devices(olm.device_store, args.filter)

    elif args.category == "verified":
        print_devices(
            olm.device_store,
            args.filter,
            "Verified",
            olm.is_device_verified
        )

    elif args.category == "unverified":
        def predicate(device):
            return not olm.is_device_verified(device)

        print_devices(
            olm.device_store,
            args.filter,
            "Unverified",
            predicate
        )

    elif args.category == "blacklisted":
        print_devices(
            olm.device_store,
            args.filter,
            "Blacklisted",
            olm.is_device_blacklisted
        )

    elif args.category == "ignored":
        print_devices(
            olm.device_store,
            args.filter,
            "Ignored",
            olm.is_device_ignored
        )


def olm_action_command(server, args, category, error_category, prefix, action):
    device_store = server.client.olm.device_store
    users = []

    if args.user_filter == "*":
        users = device_store.users
    else:
        users = [x for x in device_store.users if args.user_filter in x]

    user_devices = {
        user: device_store.active_user_devices(user) for user in users
    }

    if args.device_filter and args.device_filter != "*":
        filtered_user_devices = {}
        for user, device_list in user_devices.items():
            filtered_devices = filter(
                lambda x: args.device_filter in x.id,
                device_list
            )
            filtered_user_devices[user] = list(filtered_devices)
        user_devices = filtered_user_devices

    changed_devices = defaultdict(list)

    for user, device_list in user_devices.items():
        for device in device_list:
            if action(device):
                changed_devices[user].append(device)

    if not changed_devices:
        message = ("{prefix}matrix: No matching {error_category} devices "
                   "found.").format(
                       prefix=W.prefix("error"),
                       error_category=error_category
                   )
        W.prnt(server.server_buffer, message)
        return

    user_strings = []
    for user_id, device_list in changed_devices.items():
        device_strings = []
        message = ("  - User: {user_color}{user}{ncolor}\n").format(
                user_color=W.color("chat_nick"),
                user=user_id,
                ncolor=W.color("reset"))
        for device in device_list:
            device_strings.append(format_device(
                device.id,
                device.ed25519,
                device.display_name
            ))
        if not device_strings:
            continue

        d_string = "\n".join(device_strings)
        message += d_string
        user_strings.append(message)

    W.prnt(server.server_buffer,
           "{}matrix: {} key(s):\n".format(W.prefix("prefix"), category))
    W.prnt(server.server_buffer, "\n".join(user_strings))
    pass


def olm_verify_command(server, args):
    olm_action_command(
        server,
        args,
        "Verified",
        "unverified",
        "join",
        server.client.verify_device
    )


def olm_unverify_command(server, args):
    olm_action_command(
        server,
        args,
        "Unverified",
        "verified",
        "quit",
        server.client.unverify_device
    )


def olm_blacklist_command(server, args):
    olm_action_command(
        server,
        args,
        "Blacklisted",
        "unblacklisted",
        "join",
        server.client.blacklist_device
    )


def olm_unblacklist_command(server, args):
    olm_action_command(
        server,
        args,
        "Unblacklisted",
        "blacklisted",
        "join",
        server.client.unblacklist_device
    )


def olm_ignore_command(server, args):
    olm_action_command(
        server,
        args,
        "Ignored",
        "ignored",
        "join",
        server.client.ignore_device
    )


def olm_unignore_command(server, args):
    olm_action_command(
        server,
        args,
        "Unignored",
        "unignored",
        "join",
        server.client.unignore_device
    )


def olm_export_command(server, args):
    file_path = os.path.expanduser(args.file)
    try:
        server.client.export_keys(file_path, args.passphrase)
        server.info("Successfully exported keys")
    except (OSError, IOError) as e:
        server.error("Error exporting keys: {}".format(str(e)))

def olm_import_command(server, args):
    file_path = os.path.expanduser(args.file)
    try:
        server.client.import_keys(file_path, args.passphrase)
        server.info("Successfully imported keys")
    except (OSError, IOError, EncryptionError) as e:
        server.error("Error importing keys: {}".format(str(e)))


def olm_sas_command(server, args):
    try:
        device_store = server.client.device_store
    except LocalProtocolError:
        server.error("The device store is not loaded")
        return W.WEECHAT_RC_OK

    try:
        device = device_store[args.user_id][args.device_id]
    except KeyError:
        server.error("Device {} of user {} not found".format(
            args.device_id,
            args.user_id
        ))
        return W.WEECHAT_RC_OK

    if device.deleted:
        server.error("Device {} of user {} is deleted.".format(
            args.device_id,
            args.user_id
        ))
        return W.WEECHAT_RC_OK

    if args.action == "start":
        server.start_verification(device)
    elif args.action in ["accept", "confirm", "cancel"]:
        sas = server.client.get_active_sas(args.user_id, args.device_id)

        if not sas:
            server.error("No active key verification found for "
                         "device {} of user {}.".format(
                             args.device_id,
                             args.user_id
                         ))
            return W.WEECHAT_RC_OK

        try:
            if args.action == "accept":
                server.accept_sas(sas)
            elif args.action == "confirm":
                server.confirm_sas(sas)
            elif args.action == "cancel":
                server.cancel_sas(sas)

        except LocalProtocolError as e:
            server.error(str(e))


@utf8_decode
def matrix_olm_command_cb(data, buffer, args):
    def command(server, data, buffer, args):
        parsed_args = WeechatCommandParser.olm(args)
        if not parsed_args:
            return W.WEECHAT_RC_OK

        if not server.client.olm:
            W.prnt(server.server_buffer, "{}matrix: Olm account isn't "
                   "loaded.".format(W.prefix("error")))
            return W.WEECHAT_RC_OK

        if not parsed_args.subcommand or parsed_args.subcommand == "info":
            olm_info_command(server, parsed_args)
        elif parsed_args.subcommand == "export":
            olm_export_command(server, parsed_args)
        elif parsed_args.subcommand == "import":
            olm_import_command(server, parsed_args)
        elif parsed_args.subcommand == "verify":
            olm_verify_command(server, parsed_args)
        elif parsed_args.subcommand == "unverify":
            olm_unverify_command(server, parsed_args)
        elif parsed_args.subcommand == "blacklist":
            olm_blacklist_command(server, parsed_args)
        elif parsed_args.subcommand == "unblacklist":
            olm_unblacklist_command(server, parsed_args)
        elif parsed_args.subcommand == "verification":
            olm_sas_command(server, parsed_args)
        elif parsed_args.subcommand == "ignore":
            olm_ignore_command(server, parsed_args)
        elif parsed_args.subcommand == "unignore":
            olm_unignore_command(server, parsed_args)
        else:
            message = ("{prefix}matrix: Command not implemented.".format(
                prefix=W.prefix("error")))
            W.prnt(server.server_buffer, message)

        W.bar_item_update("buffer_modes")
        W.bar_item_update("matrix_modes")

        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer in server.buffers.values():
            return command(server, data, buffer, args)
        elif buffer == server.server_buffer:
            return command(server, data, buffer, args)

    W.prnt("", "{prefix}matrix: command \"olm\" must be executed on a "
           "matrix buffer (server or channel)".format(
               prefix=W.prefix("error")
           ))

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_devices_command_cb(data, buffer, args):
    for server in SERVERS.values():
        if buffer in server.buffers.values() or buffer == server.server_buffer:
            parsed_args = WeechatCommandParser.devices(args)
            if not parsed_args:
                return W.WEECHAT_RC_OK

            if not parsed_args.subcommand or parsed_args.subcommand == "list":
                server.devices()
            elif parsed_args.subcommand == "delete":
                server.delete_device(parsed_args.device_id)
            elif parsed_args.subcommand == "set-name":
                new_name = " ".join(parsed_args.device_name).strip("\"")
                server.rename_device(parsed_args.device_id, new_name)

            return W.WEECHAT_RC_OK

    W.prnt("", "{prefix}matrix: command \"devices\" must be executed on a "
           "matrix buffer (server or channel)".format(
               prefix=W.prefix("error")
           ))

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_me_command_cb(data, buffer, args):
    for server in SERVERS.values():
        if buffer in server.buffers.values():

            if not server.connected:
                message = (
                    "{prefix}matrix: you are not connected to " "the server"
                ).format(prefix=W.prefix("error"))
                W.prnt(server.server_buffer, message)
                return W.WEECHAT_RC_ERROR

            room_buffer = server.find_room_from_ptr(buffer)

            if not server.client.logged_in:
                room_buffer.error("You are not logged in.")
                return W.WEECHAT_RC_ERROR

            if not args:
                return W.WEECHAT_RC_OK

            formatted_data = Formatted.from_input_line(args)

            server.room_send_message(room_buffer, formatted_data, "m.emote")
            return W.WEECHAT_RC_OK

        if buffer == server.server_buffer:
            message = (
                '{prefix}matrix: command "me" must be '
                "executed on a Matrix channel buffer"
            ).format(prefix=W.prefix("error"))
            W.prnt("", message)
            return W.WEECHAT_RC_OK

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_topic_command_cb(data, buffer, args):
    parsed_args = WeechatCommandParser.topic(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer == server.server_buffer:
            server.error(
                'command "topic" must be ' "executed on a Matrix room buffer"
            )
            return W.WEECHAT_RC_OK

        room = server.find_room_from_ptr(buffer)
        if not room:
            continue

        if not parsed_args.topic and not parsed_args.delete:
            # TODO print the current topic
            return W.WEECHAT_RC_OK

        if parsed_args.delete and parsed_args.topic:
            # TODO error message
            return W.WEECHAT_RC_OK

        topic = "" if parsed_args.delete else " ".join(parsed_args.topic)
        content = {"topic": topic}
        server.room_send_state(room, content, "m.room.topic")

        return W.WEECHAT_RC_OK


def matrix_fetch_old_messages(server, room_id):
    room_buffer = server.find_room_from_id(room_id)
    room = room_buffer.room

    if room_buffer.backlog_pending:
        return

    prev_batch = room.prev_batch

    if not prev_batch:
        return

    raise NotImplementedError


def check_server_existence(server_name, servers):
    if server_name not in servers:
        message = "{prefix}matrix: No such server: {server}".format(
            prefix=W.prefix("error"), server=server_name
        )
        W.prnt("", message)
        return False
    return True


def hook_page_up():
    G.CONFIG.page_up_hook = W.hook_command_run(
        "/window page_up", "matrix_command_pgup_cb", ""
    )


@utf8_decode
def matrix_command_buf_clear_cb(data, buffer, command):
    for server in SERVERS.values():
        if buffer in server.buffers.values():
            room_buffer = server.find_room_from_ptr(buffer)
            room_buffer.room.prev_batch = server.next_batch

            return W.WEECHAT_RC_OK

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_command_pgup_cb(data, buffer, command):
    # TODO the highlight status of a line isn't allowed to be updated/changed
    # via hdata, therefore the highlight status of a messages can't be
    # reoredered this would need to be fixed in weechat
    # TODO we shouldn't fetch and print out more messages than
    # max_buffer_lines_number or older messages than max_buffer_lines_minutes
    for server in SERVERS.values():
        if buffer in server.buffers.values():
            window = W.window_search_with_buffer(buffer)

            first_line_displayed = bool(
                W.window_get_integer(window, "first_line_displayed")
            )

            room_buffer = server.find_room_from_ptr(buffer)

            if first_line_displayed or room_buffer.weechat_buffer.num_lines == 0:
                room_id = key_from_value(server.buffers, buffer)
                server.room_get_messages(room_id)

            return W.WEECHAT_RC_OK

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_join_command_cb(data, buffer, args):
    parsed_args = WeechatCommandParser.join(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer in server.buffers.values() or buffer == server.server_buffer:
            server.room_join(parsed_args.room_id)
            break

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_part_command_cb(data, buffer, args):
    parsed_args = WeechatCommandParser.part(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer in server.buffers.values() or buffer == server.server_buffer:
            room_id = parsed_args.room_id

            if not room_id:
                if buffer == server.server_buffer:
                    server.error(
                        'command "part" must be '
                        "executed on a Matrix room buffer or a room "
                        "name needs to be given"
                    )
                    return W.WEECHAT_RC_OK

                room_buffer = server.find_room_from_ptr(buffer)
                room_id = room_buffer.room.room_id

            server.room_leave(room_id)
            break

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_invite_command_cb(data, buffer, args):
    parsed_args = WeechatCommandParser.invite(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer == server.server_buffer:
            server.error(
                'command "invite" must be ' "executed on a Matrix room buffer"
            )
            return W.WEECHAT_RC_OK

        room = server.find_room_from_ptr(buffer)
        if not room:
            continue

        user_id = parsed_args.user_id
        user_id = user_id if user_id.startswith("@") else "@" + user_id

        server.room_invite(room, user_id)
        break

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_room_command_cb(data, buffer, args):
    parsed_args = WeechatCommandParser.room(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer == server.server_buffer:
            server.error(
                'command "room" must be ' "executed on a Matrix room buffer"
            )
            return W.WEECHAT_RC_OK

        room = server.find_room_from_ptr(buffer)
        if not room:
            continue

        if not parsed_args.subcommand or parsed_args.subcommand == "list":
            server.error("command no subcommand found")
            return W.WEECHAT_RC_OK

        if parsed_args.subcommand == "typing-notifications":
            if parsed_args.state == "enable":
                room.typing_enabled = True
            elif parsed_args.state == "disable":
                room.typing_enabled = False
            elif parsed_args.state == "toggle":
                room.typing_enabled = not room.typing_enabled
            break

        elif parsed_args.subcommand == "read-markers":
            if parsed_args.state == "enable":
                room.read_markers_enabled = True
            elif parsed_args.state == "disable":
                room.read_markers_enabled = False
            elif parsed_args.state == "toggle":
                room.read_markers_enabled = not room.read_markers_enabled
            break

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_uploads_command_cb(data, buffer, args):
    if not args:
        if not G.CONFIG.upload_buffer:
            G.CONFIG.upload_buffer = UploadsBuffer()
        G.CONFIG.upload_buffer.display()
        return W.WEECHAT_RC_OK

    parsed_args = WeechatCommandParser.uploads(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    if parsed_args.subcommand == "list":
        pass
    elif parsed_args.subcommand == "listfull":
        pass
    elif parsed_args.subcommand == "up":
        if G.CONFIG.upload_buffer:
            G.CONFIG.upload_buffer.move_line_up()
    elif parsed_args.subcommand == "down":
        if G.CONFIG.upload_buffer:
            G.CONFIG.upload_buffer.move_line_down()

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_upload_command_cb(data, buffer, args):
    parsed_args = WeechatCommandParser.upload(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer == server.server_buffer:
            server.error(
                'command "upload" must be ' "executed on a Matrix room buffer"
            )
            return W.WEECHAT_RC_OK

        room_buffer = server.find_room_from_ptr(buffer)
        if not room_buffer:
            continue

        upload = Upload(
            server.name,
            server.config.address,
            server.client.access_token,
            room_buffer.room.room_id,
            parsed_args.file,
            room_buffer.room.encrypted
        )
        UPLOADS[upload.uuid] = upload

        if G.CONFIG.upload_buffer:
            G.CONFIG.upload_buffer.render()

        break

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_kick_command_cb(data, buffer, args):
    parsed_args = WeechatCommandParser.kick(args)
    if not parsed_args:
        return W.WEECHAT_RC_OK

    for server in SERVERS.values():
        if buffer == server.server_buffer:
            server.error(
                'command "kick" must be ' "executed on a Matrix room buffer"
            )
            return W.WEECHAT_RC_OK

        room = server.find_room_from_ptr(buffer)
        if not room:
            continue

        user_id = parsed_args.user_id
        user_id = user_id if user_id.startswith("@") else "@" + user_id
        reason = " ".join(parsed_args.reason) if parsed_args.reason else None

        server.room_kick(room, user_id, reason)
        break

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_redact_command_cb(data, buffer, args):
    def already_redacted(line):
        if SCRIPT_NAME + "_redacted" in line.tags:
            return True
        return False

    def predicate(event_id, line):
        event_tag = SCRIPT_NAME + "_id_{}".format(event_id)
        tags = line.tags

        if event_tag in tags:
            return True

        return False

    for server in SERVERS.values():
        if buffer in server.buffers.values():
            room_buffer = server.find_room_from_ptr(buffer)

            event_id, reason = parse_redact_args(args)

            if not event_id:
                message = (
                    "{prefix}matrix: Invalid command "
                    "arguments (see /help redact)"
                ).format(prefix=W.prefix("error"))
                W.prnt("", message)
                return W.WEECHAT_RC_ERROR

            lines = room_buffer.weechat_buffer.find_lines(
                partial(predicate, event_id), max_lines=1
            )

            if not lines:
                room_buffer.error(
                    "No such message with event id "
                    "{event_id} found.".format(event_id=event_id))
                return W.WEECHAT_RC_OK

            if already_redacted(lines[0]):
                room_buffer.error("Event already redacted.")
                return W.WEECHAT_RC_OK

            server.room_send_redaction(room_buffer, event_id, reason)

            return W.WEECHAT_RC_OK

        if buffer == server.server_buffer:
            message = (
                '{prefix}matrix: command "redact" must be '
                "executed on a Matrix channel buffer"
            ).format(prefix=W.prefix("error"))
            W.prnt("", message)
            return W.WEECHAT_RC_OK

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_reply_command_cb(data, buffer, args):
    def predicate(event_id, line):
        event_tag = SCRIPT_NAME + "_id_{}".format(event_id)
        tags = line.tags

        if event_tag in tags:
            return True
        return False

    for server in SERVERS.values():
        if buffer in server.buffers.values():
            room_buffer = server.find_room_from_ptr(buffer)

            # Intentional use of `parse_redact_args` which serves the
            # necessary purpose
            event_id, reply = parse_redact_args(args)

            if not event_id or not reply:
                message = (
                    "{prefix}matrix: Invalid command "
                    "arguments (see /help reply)"
                ).format(prefix=W.prefix("error"))
                W.prnt("", message)
                return W.WEECHAT_RC_ERROR

            lines = room_buffer.weechat_buffer.find_lines(
                partial(predicate, event_id), max_lines=1
            )

            if not lines:
                room_buffer.error(
                    "No such message with event id "
                    "{event_id} found.".format(event_id=event_id))
                return W.WEECHAT_RC_OK

            formatted_data = Formatted.from_input_line(reply)
            server.room_send_message(
                room_buffer,
                formatted_data,
                "m.text",
                in_reply_to_event_id=event_id,
            )
            room_buffer.last_message = None

            return W.WEECHAT_RC_OK

        if buffer == server.server_buffer:
            message = (
                '{prefix}matrix: command "reply" must be '
                "executed on a Matrix channel buffer"
            ).format(prefix=W.prefix("error"))
            W.prnt("", message)
            return W.WEECHAT_RC_OK

    return W.WEECHAT_RC_OK


def matrix_command_help(args):
    if not args:
        message = (
            "{prefix}matrix: Too few arguments for command "
            '"/matrix help" (see /matrix help help)'
        ).format(prefix=W.prefix("error"))
        W.prnt("", message)
        return

    for command in args:
        message = ""

        if command == "connect":
            message = (
                "{delimiter_color}[{ncolor}matrix{delimiter_color}]  "
                "{ncolor}{cmd_color}/connect{ncolor} "
                "<server-name> [<server-name>...]"
                "\n\n"
                "connect to Matrix server(s)"
                "\n\n"
                "server-name: server to connect to"
                "(internal name)"
            ).format(
                delimiter_color=W.color("chat_delimiters"),
                cmd_color=W.color("chat_buffer"),
                ncolor=W.color("reset"),
            )

        elif command == "disconnect":
            message = (
                "{delimiter_color}[{ncolor}matrix{delimiter_color}]  "
                "{ncolor}{cmd_color}/disconnect{ncolor} "
                "<server-name> [<server-name>...]"
                "\n\n"
                "disconnect from Matrix server(s)"
                "\n\n"
                "server-name: server to disconnect"
                "(internal name)"
            ).format(
                delimiter_color=W.color("chat_delimiters"),
                cmd_color=W.color("chat_buffer"),
                ncolor=W.color("reset"),
            )

        elif command == "reconnect":
            message = (
                "{delimiter_color}[{ncolor}matrix{delimiter_color}]  "
                "{ncolor}{cmd_color}/reconnect{ncolor} "
                "<server-name> [<server-name>...]"
                "\n\n"
                "reconnect to Matrix server(s)"
                "\n\n"
                "server-name: server to reconnect"
                "(internal name)"
            ).format(
                delimiter_color=W.color("chat_delimiters"),
                cmd_color=W.color("chat_buffer"),
                ncolor=W.color("reset"),
            )

        elif command == "server":
            message = (
                "{delimiter_color}[{ncolor}matrix{delimiter_color}]  "
                "{ncolor}{cmd_color}/server{ncolor} "
                "add <server-name> <hostname>[:<port>]"
                "\n                  "
                "delete|list|listfull <server-name>"
                "\n\n"
                "list, add, or remove Matrix servers"
                "\n\n"
                "       list: list servers (without argument, this "
                "list is displayed)\n"
                "   listfull: list servers with detailed info for each "
                "server\n"
                "        add: add a new server\n"
                "     delete: delete a server\n"
                "server-name: server to reconnect (internal name)\n"
                "   hostname: name or IP address of server\n"
                "       port: port of server (default: 443)\n"
                "\n"
                "Examples:"
                "\n  /matrix server listfull"
                "\n  /matrix server add matrix matrix.org:80"
                "\n  /matrix server delete matrix"
            ).format(
                delimiter_color=W.color("chat_delimiters"),
                cmd_color=W.color("chat_buffer"),
                ncolor=W.color("reset"),
            )

        elif command == "help":
            message = (
                "{delimiter_color}[{ncolor}matrix{delimiter_color}]  "
                "{ncolor}{cmd_color}/help{ncolor} "
                "<matrix-command> [<matrix-command>...]"
                "\n\n"
                "display help about Matrix commands"
                "\n\n"
                "matrix-command: a Matrix command name"
                "(internal name)"
            ).format(
                delimiter_color=W.color("chat_delimiters"),
                cmd_color=W.color("chat_buffer"),
                ncolor=W.color("reset"),
            )

        else:
            message = (
                '{prefix}matrix: No help available, "{command}" '
                "is not a matrix command"
            ).format(prefix=W.prefix("error"), command=command)

        W.prnt("", "")
        W.prnt("", message)

        return


def matrix_server_command_listfull(args):
    def get_value_string(value, default_value):
        if value == default_value:
            if not value:
                value = "''"
            value_string = "  ({value})".format(value=value)
        else:
            value_string = "{color}{value}{ncolor}".format(
                color=W.color("chat_value"),
                value=value,
                ncolor=W.color("reset"),
            )

        return value_string

    for server_name in args:
        if server_name not in SERVERS:
            continue

        server = SERVERS[server_name]
        connected = ""

        W.prnt("", "")

        if server.connected:
            connected = "connected"
        else:
            connected = "not connected"

        message = (
            "Server: {server_color}{server}{delimiter_color}"
            " [{ncolor}{connected}{delimiter_color}]"
            "{ncolor}"
        ).format(
            server_color=W.color("chat_server"),
            server=server.name,
            delimiter_color=W.color("chat_delimiters"),
            connected=connected,
            ncolor=W.color("reset"),
        )

        W.prnt("", message)

        option = server.config._option_ptrs["autoconnect"]
        default_value = W.config_string_default(option)
        value = W.config_string(option)

        value_string = get_value_string(value, default_value)
        message = "  autoconnect. : {value}".format(value=value_string)

        W.prnt("", message)

        option = server.config._option_ptrs["address"]
        default_value = W.config_string_default(option)
        value = W.config_string(option)

        value_string = get_value_string(value, default_value)
        message = "  address. . . : {value}".format(value=value_string)

        W.prnt("", message)

        option = server.config._option_ptrs["port"]
        default_value = str(W.config_integer_default(option))
        value = str(W.config_integer(option))

        value_string = get_value_string(value, default_value)
        message = "  port . . . . : {value}".format(value=value_string)

        W.prnt("", message)

        option = server.config._option_ptrs["username"]
        default_value = W.config_string_default(option)
        value = W.config_string(option)

        value_string = get_value_string(value, default_value)
        message = "  username . . : {value}".format(value=value_string)

        W.prnt("", message)

        option = server.config._option_ptrs["password"]
        value = W.config_string(option)

        if value:
            value = "(hidden)"

        value_string = get_value_string(value, "")
        message = "  password . . : {value}".format(value=value_string)

        W.prnt("", message)


def matrix_server_command_delete(args):
    for server_name in args:
        if check_server_existence(server_name, SERVERS):
            server = SERVERS[server_name]

            if server.connected:
                message = (
                    "{prefix}matrix: you can not delete server "
                    "{color}{server}{ncolor} because you are "
                    'connected to it. Try "/matrix disconnect '
                    '{color}{server}{ncolor}" before.'
                ).format(
                    prefix=W.prefix("error"),
                    color=W.color("chat_server"),
                    ncolor=W.color("reset"),
                    server=server.name,
                )
                W.prnt("", message)
                return

            for buf in list(server.buffers.values()):
                W.buffer_close(buf)

            if server.server_buffer:
                W.buffer_close(server.server_buffer)

            for option in server.config._option_ptrs.values():
                W.config_option_free(option)

            if server.timer_hook:
                W.unhook(server.timer_hook)
                server.timer_hook = None

            message = (
                "matrix: server {color}{server}{ncolor} has been " "deleted"
            ).format(
                server=server.name,
                color=W.color("chat_server"),
                ncolor=W.color("reset"),
            )

            del SERVERS[server.name]
            server = None

            W.prnt("", message)


def matrix_server_command_add(args):
    if len(args) < 2:
        message = (
            "{prefix}matrix: Too few arguments for command "
            '"/matrix server add" (see /matrix help server)'
        ).format(prefix=W.prefix("error"))
        W.prnt("", message)
        return
    if len(args) > 4:
        message = (
            "{prefix}matrix: Too many arguments for command "
            '"/matrix server add" (see /matrix help server)'
        ).format(prefix=W.prefix("error"))
        W.prnt("", message)
        return

    def remove_server(server):
        for option in server.config._option_ptrs.values():
            W.config_option_free(option)
        del SERVERS[server.name]

    server_name = args[0]

    if server_name in SERVERS:
        message = (
            "{prefix}matrix: server {color}{server}{ncolor} "
            "already exists, can't add it"
        ).format(
            prefix=W.prefix("error"),
            color=W.color("chat_server"),
            server=server_name,
            ncolor=W.color("reset"),
        )
        W.prnt("", message)
        return

    server = MatrixServer(server_name, G.CONFIG._ptr)
    SERVERS[server.name] = server

    if len(args) >= 2:
        if args[1].startswith("http"):
            homeserver= urlparse(args[1])

            host = homeserver.hostname
            port = str(homeserver.port) if homeserver.port else None
        else:
            try:
                host, port = args[1].split(":", 1)
            except ValueError:
                host, port = args[1], None

        return_code = W.config_option_set(
            server.config._option_ptrs["address"], host, 1
        )

        if return_code == W.WEECHAT_CONFIG_OPTION_SET_ERROR:
            remove_server(server)
            message = (
                "{prefix}Failed to set address for server "
                "{color}{server}{ncolor}, failed to add "
                "server."
            ).format(
                prefix=W.prefix("error"),
                color=W.color("chat_server"),
                server=server.name,
                ncolor=W.color("reset"),
            )

            W.prnt("", message)
            server = None
            return

        if port:
            return_code = W.config_option_set(
                server.config._option_ptrs["port"], port, 1
            )
            if return_code == W.WEECHAT_CONFIG_OPTION_SET_ERROR:
                remove_server(server)
                message = (
                    "{prefix}Failed to set port for server "
                    "{color}{server}{ncolor}, failed to add "
                    "server."
                ).format(
                    prefix=W.prefix("error"),
                    color=W.color("chat_server"),
                    server=server.name,
                    ncolor=W.color("reset"),
                )

                W.prnt("", message)
                server = None
                return

    if len(args) >= 3:
        user = args[2]
        return_code = W.config_option_set(
            server.config._option_ptrs["username"], user, 1
        )

        if return_code == W.WEECHAT_CONFIG_OPTION_SET_ERROR:
            remove_server(server)
            message = (
                "{prefix}Failed to set user for server "
                "{color}{server}{ncolor}, failed to add "
                "server."
            ).format(
                prefix=W.prefix("error"),
                color=W.color("chat_server"),
                server=server.name,
                ncolor=W.color("reset"),
            )

            W.prnt("", message)
            server = None
            return

    if len(args) == 4:
        password = args[3]

        return_code = W.config_option_set(
            server.config._option_ptrs["password"], password, 1
        )
        if return_code == W.WEECHAT_CONFIG_OPTION_SET_ERROR:
            remove_server(server)
            message = (
                "{prefix}Failed to set password for server "
                "{color}{server}{ncolor}, failed to add "
                "server."
            ).format(
                prefix=W.prefix("error"),
                color=W.color("chat_server"),
                server=server.name,
                ncolor=W.color("reset"),
            )
            W.prnt("", message)
            server = None
            return

    message = (
        "matrix: server {color}{server}{ncolor} " "has been added"
    ).format(
        server=server.name,
        color=W.color("chat_server"),
        ncolor=W.color("reset"),
    )
    W.prnt("", message)


def matrix_server_command(command, args):
    def list_servers(_):
        if SERVERS:
            W.prnt("", "\nAll matrix servers:")
            for server in SERVERS:
                W.prnt(
                    "",
                    "    {color}{server}".format(
                        color=W.color("chat_server"), server=server
                    ),
                )

    # TODO the argument for list and listfull is used as a match word to
    # find/filter servers, we're currently match exactly to the whole name
    if command == "list":
        list_servers(args)
    elif command == "listfull":
        matrix_server_command_listfull(args)
    elif command == "add":
        matrix_server_command_add(args)
    elif command == "delete":
        matrix_server_command_delete(args)
    else:
        message = (
            "{prefix}matrix: Error: unknown matrix server command, "
            '"{command}" (type /matrix help server for help)'
        ).format(prefix=W.prefix("error"), command=command)
        W.prnt("", message)


@utf8_decode
def matrix_command_cb(data, buffer, args):
    def connect_server(args):
        for server_name in args:
            if check_server_existence(server_name, SERVERS):
                server = SERVERS[server_name]
                server.connect()

    def disconnect_server(args):
        for server_name in args:
            if check_server_existence(server_name, SERVERS):
                server = SERVERS[server_name]
                if server.connected or server.reconnect_time:
                    # W.unhook(server.timer_hook)
                    # server.timer_hook = None
                    server.access_token = ""
                    server.disconnect(reconnect=False)

    split_args = list(filter(bool, args.split(" ")))

    if len(split_args) < 1:
        message = (
            "{prefix}matrix: Too few arguments for command "
            '"/matrix" '
            "(see /help matrix)"
        ).format(prefix=W.prefix("error"))
        W.prnt("", message)
        return W.WEECHAT_RC_ERROR

    command, args = split_args[0], split_args[1:]

    if command == "connect":
        connect_server(args)

    elif command == "disconnect":
        disconnect_server(args)

    elif command == "reconnect":
        disconnect_server(args)
        connect_server(args)

    elif command == "server":
        if len(args) >= 1:
            subcommand, args = args[0], args[1:]
            matrix_server_command(subcommand, args)
        else:
            matrix_server_command("list", "")

    elif command == "help":
        matrix_command_help(args)

    else:
        message = (
            "{prefix}matrix: Error: unknown matrix command, "
            '"{command}" (type /help matrix for help)'
        ).format(prefix=W.prefix("error"), command=command)
        W.prnt("", message)

    return W.WEECHAT_RC_OK


@utf8_decode
def matrix_send_anyways_cb(data, buffer, args):
    for server in SERVERS.values():
        if buffer in server.buffers.values():
            room_buffer = server.find_room_from_ptr(buffer)

            if not server.connected:
                room_buffer.error("Server is disconnected")
                break

            if not server.client.logged_in:
                room_buffer.error("You are not logged in.")
                return W.WEECHAT_RC_ERROR

            if not room_buffer.last_message:
                room_buffer.error("No previously sent message found.")
                break

            server.room_send_message(
                room_buffer,
                room_buffer.last_message,
                "m.text",
                ignore_unverified_devices=True
            )
            room_buffer.last_message = None

            break
    else:
        message = (
            "{prefix}matrix: The 'send-anyways' command needs to be "
            "run on a matrix room buffer"
        ).format(prefix=W.prefix("error"))
        W.prnt("", message)

    return W.WEECHAT_RC_ERROR


@utf8_decode
def matrix_cursor_reply_signal_cb(data, signal, ht):
    tags = ht["_chat_line_tags"].split(",")

    W.command("", "/cursor stop")

    if "matrix_message" in tags:
        for tag in tags:
            if tag.startswith("matrix_id_"):
                matrix_id = tag[10:]
                break
        else:
            return W.WEECHAT_RC_OK

        buffer_name = ht["_buffer_full_name"]
        bufptr = W.buffer_search("==", buffer_name)

        current_input = W.buffer_get_string(bufptr, "input")
        input_pos = W.buffer_get_integer(bufptr, "input_pos")

        new_prefix = "/reply-matrix {} ".format(matrix_id)

        W.buffer_set(bufptr, "input", new_prefix + current_input)
        W.buffer_set(bufptr, "input_pos", str(len(new_prefix) + input_pos))

    return W.WEECHAT_RC_OK



import datetime
import random
import string

WEECHAT_BASE_COLORS = {
    "black":        "0",
    "red":          "1",
    "green":        "2",
    "brown":        "3",
    "blue":         "4",
    "magenta":      "5",
    "cyan":         "6",
    "default":      "7",
    "gray":         "8",
    "lightred":     "9",
    "lightgreen":   "10",
    "yellow":       "11",
    "lightblue":    "12",
    "lightmagenta": "13",
    "lightcyan":    "14",
    "white":        "15"
}


class MockObject(object):
    pass

class MockConfig(object):
    config_template = {
        'debug_buffer': None,
        'debug_category': None,
        '_ptr': None,
        'read': None,
        'free': None,
        'page_up_hook': None,
        'color': {
            'error_message_bg': "",
            'error_message_fg': "",
            'quote_bg': "",
            'quote_fg': "",
            'unconfirmed_message_bg': "",
            'unconfirmed_message_fg': "",
            'untagged_code_bg': "",
            'untagged_code_fg': "",
        },
        'upload_buffer': {
            'display': None,
            'move_line_down': None,
            'move_line_up': None,
            'render': None,
        },
        'look': {
            'bar_item_typing_notice_prefix': None,
            'busy_sign': None,
            'code_block_margin': None,
            'code_blocks': None,
            'disconnect_sign': None,
            'encrypted_room_sign': None,
            'encryption_warning_sign': None,
            'max_typing_notice_item_length': None,
            'pygments_style': None,
            'redactions': None,
            'server_buffer': None,
            'new_channel_position': None,
            'markdown_input': True,
        },
        'network': {
            'debug_buffer': None,
            'debug_category': None,
            'debug_level': None,
            'fetch_backlog_on_pgup': None,
            'lag_min_show': None,
            'lag_reconnect': None,
            'lazy_load_room_users': None,
            'max_initial_sync_events': None,
            'max_nicklist_users': None,
            'print_unconfirmed_messages': None,
            'read_markers_conditions': None,
            'typing_notice_conditions': None,
            'autoreconnect_delay_growing': None,
            'autoreconnect_delay_max': None,
        },
    }

    def __init__(self):
        for category, options in MockConfig.config_template.items():
            if options:
                category_object = MockObject()
                for option, value in options.items():
                    setattr(category_object, option, value)
            else:
                category_object = options

            setattr(self, category, category_object)


def color(color_name):
    # type: (str) -> str
    # yapf: disable
    escape_codes = []
    reset_code = "0"

    def make_fg_color(color_code):
        return "38;5;{}".format(color_code)

    def make_bg_color(color_code):
        return "48;5;{}".format(color_code)

    attributes = {
        "bold":       "1",
        "-bold":      "21",
        "reverse":    "27",
        "-reverse":   "21",
        "italic":     "3",
        "-italic":    "23",
        "underline":  "4",
        "-underline": "24",
        "reset":      "0",
        "resetcolor": "39"
    }

    short_attributes = {
        "*": "1",
        "!": "27",
        "/": "3",
        "_": "4"
    }

    colors = color_name.split(",", 2)

    fg_color = colors.pop(0)

    bg_color = colors.pop(0) if colors else ""

    if fg_color in attributes:
        escape_codes.append(attributes[fg_color])
    else:
        chars = list(fg_color)

        for char in chars:
            if char in short_attributes:
                escape_codes.append(short_attributes[char])
            elif char == "|":
                reset_code = ""
            else:
                break

        stripped_color = fg_color.lstrip("*_|/!")

        if stripped_color in WEECHAT_BASE_COLORS:
            escape_codes.append(
                make_fg_color(WEECHAT_BASE_COLORS[stripped_color]))

        elif stripped_color.isdigit():
            num_color = int(stripped_color)
            if 0 <= num_color < 256:
                escape_codes.append(make_fg_color(stripped_color))

    if bg_color in WEECHAT_BASE_COLORS:
        escape_codes.append(make_bg_color(WEECHAT_BASE_COLORS[bg_color]))
    else:
        if bg_color.isdigit():
            num_color = int(bg_color)
            if 0 <= num_color < 256:
                escape_codes.append(make_bg_color(bg_color))

    escape_string = "\033[{}{}m".format(reset_code, ";".join(escape_codes))

    return escape_string


def prefix(prefix_string):
    prefix_to_symbol = {
        "error":   "=!=",
        "network": "--",
        "action":  "*",
        "join":    "-->",
        "quit":    "<--"
    }

    if prefix_string in prefix_to_symbol:
        return prefix_to_symbol[prefix]

    return ""


def prnt(_, message):
    print(message)


def prnt_date_tags(_, date, tags_string, data):
    message = "{} {} [{}]".format(
        datetime.datetime.fromtimestamp(date),
        data,
        tags_string
    )
    print(message)


def config_search_section(*_, **__):
    pass


def config_new_option(*_, **__):
    pass


def mkdir_home(*_, **__):
    return True


def info_get(info, *_):
    if info == "nick_color_name":
        return random.choice(list(WEECHAT_BASE_COLORS.keys()))

    return ""


def buffer_new(*_, **__):
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(8)
    )


def buffer_set(*_, **__):
    return


def buffer_get_string(_ptr, property):
    if property == "localvar_type":
        return "channel"
    return ""


def buffer_get_integer(_ptr, property):
    return 0


def current_buffer():
    return 1


def nicklist_add_group(*_, **__):
    return


def nicklist_add_nick(*_, **__):
    return


def nicklist_remove_nick(*_, **__):
    return


def nicklist_search_nick(*args, **kwargs):
    return buffer_new(args, kwargs)


def string_remove_color(message, _):
    return message



# -*- coding: utf-8 -*-

# Copyright (c) 2014-2016 Ryan Huber <rhuber@gmail.com>
# Copyright (c) 2015-2016 Tollef Fog Heen <tfheen@err.no>

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

import sys

# pylint: disable=redefined-builtin
from builtins import bytes, str
from functools import wraps

if sys.version_info.major == 3 and sys.version_info.minor >= 3:
    from collections.abc import Iterable, Mapping
else:
    from collections import Iterable, Mapping

# These functions were written by Trygve Aaberge for wee-slack and are under a
# MIT License.
# More info can be found in the wee-slack repository under the commit:
# 5e1c7e593d70972afb9a55f29d13adaf145d0166, the repository can be found at:
# https://github.com/wee-slack/wee-slack


class WeechatWrapper(object):
    def __init__(self, wrapped_class):
        self.wrapped_class = wrapped_class

    # Helper method used to encode/decode method calls.
    def wrap_for_utf8(self, method):
        def hooked(*args, **kwargs):
            result = method(*encode_to_utf8(args), **encode_to_utf8(kwargs))
            # Prevent wrapped_class from becoming unwrapped
            if result == self.wrapped_class:
                return self
            return decode_from_utf8(result)

        return hooked

    # Encode and decode everything sent to/received from weechat. We use the
    # unicode type internally in wee-slack, but has to send utf8 to weechat.
    def __getattr__(self, attr):
        orig_attr = self.wrapped_class.__getattribute__(attr)
        if callable(orig_attr):
            return self.wrap_for_utf8(orig_attr)
        return decode_from_utf8(orig_attr)

    # Ensure all lines sent to weechat specify a prefix. For lines after the
    # first, we want to disable the prefix, which is done by specifying a
    # space.
    def prnt_date_tags(self, buffer, date, tags, message):
        message = message.replace("\n", "\n \t")
        return self.wrap_for_utf8(self.wrapped_class.prnt_date_tags)(
            buffer, date, tags, message
        )


def utf8_decode(function):
    """
    Decode all arguments from byte strings to unicode strings. Use this for
    functions called from outside of this script, e.g. callbacks from weechat.
    """

    @wraps(function)
    def wrapper(*args, **kwargs):

        # Don't do anything if we're python 3
        if sys.hexversion >= 0x3000000:
            return function(*args, **kwargs)

        return function(*decode_from_utf8(args), **decode_from_utf8(kwargs))

    return wrapper


def decode_from_utf8(data):
    if isinstance(data, bytes):
        return data.decode("utf-8")
    if isinstance(data, str):
        return data
    elif isinstance(data, Mapping):
        return type(data)(map(decode_from_utf8, data.items()))
    elif isinstance(data, Iterable):
        return type(data)(map(decode_from_utf8, data))
    return data


def encode_to_utf8(data):
    if isinstance(data, str):
        return data.encode("utf-8")
    if isinstance(data, bytes):
        return data
    elif isinstance(data, Mapping):
        return type(data)(map(encode_to_utf8, data.items()))
    elif isinstance(data, Iterable):
        return type(data)(map(encode_to_utf8, data))
    return data
